import math
from typing import Optional

import torch
from torch import nn
from torch.nn import Parameter
from torch.nn import functional as F
from torchvision.ops import StochasticDepth

from eir.models.layers.norm_layers import LayerScale
from eir.utils.logging import get_logger

logger = get_logger(name=__name__)


class LCL(nn.Module):
    __constants__ = ["bias", "in_features", "out_features"]

    def __init__(
        self,
        in_features: int,
        out_feature_sets: int,
        num_chunks: int = 10,
        kernel_size: Optional[int] = None,
        bias: Optional[bool] = True,
    ):
        super().__init__()

        self.in_features = in_features
        self.out_feature_sets = out_feature_sets
        self.num_chunks = num_chunks

        if kernel_size:
            self.kernel_size = kernel_size
            self.num_chunks = int(math.ceil(in_features / kernel_size))
            logger.debug(
                "%s: Setting num chunks to %d as kernel size of %d was passed in.",
                self.__class__,
                self.num_chunks,
                kernel_size,
            )
        else:
            self.kernel_size = int(math.ceil(self.in_features / self.num_chunks))
            logger.debug(
                "%s :Setting kernel size to %d as number of "
                "chunks of %d was passed in.",
                self.__class__,
                self.kernel_size,
                self.num_chunks,
            )

        assert self.kernel_size is not None

        self.out_features = self.out_feature_sets * self.num_chunks

        self.padding = _find_lcl_padding_needed(
            input_size=self.in_features,
            kernel_size=self.kernel_size,
            num_chunks=self.num_chunks,
        )

        self.weight = Parameter(
            torch.Tensor(self.out_feature_sets, self.num_chunks, self.kernel_size),
            requires_grad=True,
        )

        if bias:
            self.bias = Parameter(torch.Tensor(self.out_features), requires_grad=True)
        else:
            self.register_parameter("bias", None)
        self.reset_parameters()

    def reset_parameters(self):
        """
        NOTE: This default init actually works quite well, as compared to initializing
        for each chunk (meaning higher weights at init). In that case, the model takes
        longer to get to a good performance as it spends a while driving the weights
        down.
        """
        nn.init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        if self.bias is not None:
            nn.init.zeros_(self.bias)

    def extra_repr(self):
        return (
            "in_features={}, num_chunks={}, kernel_size={}, "
            "out_feature_sets={}, out_features={}, bias={}".format(
                self.in_features,
                self.num_chunks,
                self.kernel_size,
                self.out_feature_sets,
                self.out_features,
                self.bias is not None,
            )
        )

    def forward(self, input: torch.Tensor):
        input_padded = F.pad(input=input, pad=[0, self.padding, 0, 0])

        input_reshaped = input_padded.reshape(
            input.shape[0], 1, self.num_chunks, self.kernel_size
        )

        out = calc_lcl_forward(input=input_reshaped, weight=self.weight, bias=self.bias)
        return out


def _find_lcl_padding_needed(input_size: int, kernel_size: int, num_chunks: int):
    return num_chunks * kernel_size - input_size


def calc_lcl_forward(input: torch.Tensor, weight: torch.Tensor, bias: torch.Tensor):
    """
    n: num samples
    c: num chunks (height)
    s: kernel size (width)
    o: output sets
    """

    summed = torch.einsum("nhw, ohw -> noh", input.squeeze(1), weight)
    flattened = summed.flatten(start_dim=1)

    final = flattened
    if bias is not None:
        final = flattened + bias

    return final


class LCLResidualBlock(nn.Module):
    def __init__(
        self,
        in_features: int,
        out_feature_sets: int,
        kernel_size: int,
        dropout_p: float = 0.0,
        stochastic_depth_p: float = 0.0,
        full_preactivation: bool = False,
        reduce_both: bool = True,
    ):
        super().__init__()

        self.in_features = in_features
        self.kernel_size = kernel_size
        self.out_feature_sets = out_feature_sets

        self.dropout_p = dropout_p
        self.full_preactivation = full_preactivation
        self.reduce_both = reduce_both
        self.stochastic_depth_p = stochastic_depth_p

        self.norm_1 = nn.RMSNorm(normalized_shape=in_features)
        self.fc_1 = LCL(
            in_features=self.in_features,
            out_feature_sets=self.out_feature_sets,
            bias=True,
            kernel_size=self.kernel_size,
        )

        self.act_1 = nn.GELU()
        self.do = nn.Dropout(p=dropout_p)

        fc_2_kwargs = _get_lcl_2_kwargs(
            in_features=self.fc_1.out_features,
            out_feature_sets=self.out_feature_sets,
            bias=True,
            kernel_size=self.kernel_size,
            reduce_both=self.reduce_both,
        )
        self.fc_2 = LCL(**fc_2_kwargs)

        self.out_features = self.fc_2.out_features

        self.ls = LayerScale(dim=self.out_features, init_values=1.0)

        if in_features == self.out_features:
            self.downsample_identity = lambda x: x
        else:
            self.downsample_identity = LCL(
                in_features=self.in_features,
                out_feature_sets=1,
                bias=True,
                num_chunks=self.fc_2.out_features,
            )

        self.stochastic_depth = StochasticDepth(p=stochastic_depth_p, mode="batch")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = self.norm_1(x)

        identity = out if self.full_preactivation else x
        identity = self.downsample_identity(identity)

        out = self.fc_1(out)

        out = self.act_1(out)
        out = self.do(out)
        out = self.fc_2(out)
        out = self.ls(out)

        out = self.stochastic_depth(out)

        return out + identity


def _get_lcl_2_kwargs(
    in_features: int,
    out_feature_sets: int,
    bias: bool,
    reduce_both: bool,
    kernel_size: int,
):
    common_kwargs = {
        "in_features": in_features,
        "out_feature_sets": out_feature_sets,
        "bias": bias,
    }

    if reduce_both:
        common_kwargs["kernel_size"] = kernel_size
    else:
        num_chunks = _calculate_num_chunks_for_equal_lcl_out_features(
            in_features=in_features,
            out_feature_sets=out_feature_sets,
        )
        common_kwargs["num_chunks"] = num_chunks

    return common_kwargs


def _calculate_num_chunks_for_equal_lcl_out_features(
    in_features: int,
    out_feature_sets: int,
) -> int:
    """
    Ensure total out features are equal to in features.
    """
    return in_features // out_feature_sets



from typing import Any, Sequence, Type

from pydantic import BaseModel, create_model

from eir.setup.schemas import InputConfig, TabularInputDataConfig


def create_tabular_model(
    name: str,
    cat_columns: Sequence[str],
    con_columns: Sequence[str],
) -> Type[BaseModel]:
    fields: dict[str, Any] = {col: (str, ...) for col in cat_columns}
    fields.update({col: (float, ...) for col in con_columns})
    return create_model(name, **fields)


def create_input_model(configs: Sequence[InputConfig]) -> Type[BaseModel]:
    fields: dict[str, Any] = {}

    for config in configs:
        input_type = config.input_info.input_type
        input_type_info = config.input_type_info

        if input_type in {"sequence", "bytes", "omics", "array", "image"}:
            fields[config.input_info.input_name] = (str, ...)
        elif input_type == "tabular":
            assert isinstance(input_type_info, TabularInputDataConfig)
            tabular_model = create_tabular_model(
                name=f"{config.input_info.input_name}_Model",
                cat_columns=list(input_type_info.input_cat_columns),
                con_columns=list(input_type_info.input_con_columns),
            )
            fields[config.input_info.input_name] = (tabular_model, ...)
        else:
            raise ValueError(f"Unknown input type {input_type}")

    return create_model("DynamicInputModel", **fields)


class ResponseModel(BaseModel):
    result: list[dict[str, Any]]



import json
import logging
from contextlib import contextmanager
from typing import Generator

import websocket
from websocket._exceptions import WebSocketBadStatusException, WebSocketTimeoutException

from eir.setup.streaming_data_setup.protocol import (
    FROM_SERVER_MESSAGE_TYPES,
    PROTOCOL_VERSION,
)

logger = logging.getLogger(__name__)


def send_validation_ids(ws_url: str, valid_ids: list[str]):
    with connect_to_server(
        websocket_url=ws_url,
        protocol_version=PROTOCOL_VERSION,
        max_size=10_000_000,
    ) as ws:
        logger.info("Sending %d validation IDs to the server.", len(valid_ids))
        ws.send(
            json.dumps(
                {"type": "setValidationIds", "payload": {"validation_ids": valid_ids}}
            )
        )

        response_data = receive_with_timeout(websocket=ws)

        if response_data["type"] != "validationIdsConfirmation":
            raise ValueError(f"Unexpected response type: {response_data['type']}")
        logger.info(f"Server response: {response_data['payload']['message']}")


@contextmanager
def connect_to_server(
    websocket_url: str,
    protocol_version: str,
    max_size: int = 10_000_000,
) -> Generator[websocket.WebSocket, None, None]:
    logger.debug(f"Attempting WebSocket connection to: {websocket_url}")

    try:
        ws = websocket.create_connection(
            url=websocket_url,
            max_size=max_size,
        )
    except WebSocketBadStatusException as e:
        logger.error(f"Failed to establish WebSocket connection: {str(e)}")
        raise
    except Exception as e:
        logger.error(f"Unexpected connection error: {type(e).__name__} - {str(e)}")
        raise

    try:
        handshake_payload = {
            "type": "handshake",
            "version": protocol_version,
        }
        ws.send(json.dumps(handshake_payload))

        response_data = receive_with_timeout(websocket=ws)
        if response_data is None:
            raise ValueError("No handshake response received from server")

        if response_data["type"] != "handshake":
            raise ValueError(
                f"Expected handshake response, got: {response_data['type']}"
            )

        if response_data["version"] != protocol_version:
            raise ValueError(
                f"Protocol version mismatch. Expected: {protocol_version}, "
                f"Got: {response_data.get('version', 'unknown')}"
            )

        logger.info(f"Connected to server with protocol version {protocol_version}")
        yield ws
    except Exception as e:
        logger.error(f"Error during handshake: {str(e)}")
        raise
    finally:
        try:
            ws.close()
        except Exception as e:
            logger.warning(f"Error closing WebSocket connection: {str(e)}")


def receive_with_timeout(websocket: websocket.WebSocket, timeout: int = 30):
    websocket.settimeout(timeout=timeout)
    try:
        raw_message = websocket.recv()
        message_parsed = json.loads(raw_message)
        handle_server_message(message=message_parsed)
        return message_parsed
    except WebSocketTimeoutException:
        logger.error("Server connection timed out")
        return None
    except json.JSONDecodeError:
        logger.error("Received message was not valid JSON")
        return None


def handle_server_message(message: dict):
    if message["type"] not in FROM_SERVER_MESSAGE_TYPES:
        logger.warning(f"Received unknown message type from server: {message['type']}")



from typing import Optional

import polars as pl

from eir.utils.logging import get_logger

logger = get_logger(name=__name__)


def add_tabular_data_to_df(
    df_tabular: pl.DataFrame,
    input_df: pl.DataFrame,
    ids_to_keep: Optional[set[str]] = None,
    source_name: str = "Tabular Data",
) -> pl.DataFrame:
    logger.debug(f"Adding tabular data from {source_name}")

    if ids_to_keep is not None:
        df_tabular = df_tabular.filter(pl.col("ID").is_in(ids_to_keep))

    rename_expr = []
    for col in df_tabular.columns:
        if col == "ID":
            rename_expr.append(pl.col("ID"))
        else:
            new_name = f"{source_name}__{col}"
            rename_expr.append(pl.col(col).alias(new_name))

    df_tabular_renamed = df_tabular.select(rename_expr)

    if input_df.height == 0:
        return df_tabular_renamed
    else:
        return input_df.join(df_tabular_renamed, on="ID", how="full", coalesce=True)



from dataclasses import dataclass
from difflib import get_close_matches
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Literal,
    Optional,
    Sequence,
    Tuple,
)

import torch
from torch import nn
from torch.utils.data._utils.collate import default_collate

from eir.data_load.data_preparation_modules.imputation import (
    impute_missing_modalities_wrapper,
)
from eir.data_load.data_utils import Batch
from eir.models.meta.meta_utils import (
    al_fusion_modules,
    al_input_modules,
    al_output_modules,
    run_meta_forward,
)
from eir.models.tensor_broker.tensor_broker_fusion_layers import (
    get_fusion_layer_wrapper,
)
from eir.models.tensor_broker.tensor_broker_projection_layers import (
    get_projection_layer,
)
from eir.setup.input_setup import al_input_objects_as_dict
from eir.setup.output_setup import al_output_objects_as_dict
from eir.setup.schemas import FusionConfig, InputConfig, OutputConfig
from eir.train_utils.step_logic import (
    al_dataloader_getitem_batch,
    prepare_base_batch_default,
)
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    pass

logger = get_logger(name=__name__)


@dataclass
class CachedTensor:
    tensor: torch.Tensor
    shape: torch.Size
    layer_path: str


def prepare_example_test_batch(
    input_objects: al_input_objects_as_dict,
    output_objects: al_output_objects_as_dict,
    model: torch.nn.Module,
    device: str,
    batch_size: int = 2,
) -> Batch:
    imputed_inputs = impute_missing_modalities_wrapper(
        inputs_values={}, inputs_objects=input_objects
    )

    loader_batch: al_dataloader_getitem_batch = (imputed_inputs, {}, list())
    batch_as_list = [loader_batch] * batch_size
    loader_batch_collated = default_collate(batch=batch_as_list)

    batch = prepare_base_batch_default(
        loader_batch=loader_batch_collated,
        input_objects=input_objects,
        output_objects=output_objects,
        model=model,
        device=device,
    )

    return batch


def create_tensor_shapes_estimation_hook(
    output_shapes: Dict[str, torch.Size],
    input_shapes: Dict[str, torch.Size],
    layer_path: str,
) -> Callable:

    def hook(
        module: nn.Module,
        args: Tuple[torch.Tensor, ...],
        kwargs: Dict[str, torch.Tensor],
        output: torch.Tensor | Dict[str, torch.Tensor],
    ):
        if isinstance(output, torch.Tensor):
            output_shapes[layer_path] = output.shape[1:]
        elif isinstance(output, dict):
            cur_output_shapes = {k: v.shape[1:] for k, v in output.items()}
            for k, v in cur_output_shapes.items():
                output_shapes[f"{layer_path}.{k}"] = v

        if len(args) > 0:
            input_tensor = args[0]
        else:
            x = kwargs.get("x")
            input_key = kwargs.get("input")
            if x is not None:
                input_tensor = x
            elif input_key is not None:
                input_tensor = input_key
            else:
                raise ValueError("No input tensor found in args or kwargs.")

        if isinstance(input_tensor, torch.Tensor):
            input_shapes[layer_path] = input_tensor.shape[1:]
        elif isinstance(input_tensor, dict):
            cur_input_shapes = {k: v.shape[1:] for k, v in input_tensor.items()}
            for k, v in cur_input_shapes.items():
                input_shapes[f"{layer_path}.{k}"] = v

    return hook


def estimate_tensor_shapes(
    module_storage: nn.Module,
    fusion_to_output_mapping: Dict[str, Literal["computed", "pass-through"]],
    example_input_data: Dict[str, torch.Tensor],
) -> Tuple[Dict[str, torch.Size], Dict[str, torch.Size]]:
    output_shapes: Dict[str, torch.Size] = {}
    input_shapes: Dict[str, torch.Size] = {}

    hooks = []

    for name, module in module_storage.named_modules():
        cur_hook = create_tensor_shapes_estimation_hook(
            output_shapes=output_shapes,
            input_shapes=input_shapes,
            layer_path=name,
        )
        hooks.append(module.register_forward_hook(cur_hook, with_kwargs=True))

    with torch.no_grad():
        run_meta_forward(
            input_modules=module_storage.input_modules,
            fusion_modules=module_storage.fusion_modules,
            output_modules=module_storage.output_modules,
            fusion_to_output_mapping=fusion_to_output_mapping,
            inputs=example_input_data,
        )

    for h in hooks:
        h.remove()

    return output_shapes, input_shapes


class ModuleStorage(nn.Module):
    def __init__(
        self,
        input_modules: al_input_modules,
        fusion_modules: al_fusion_modules,
        output_modules: al_output_modules,
    ):
        """
        We use this to imitate the future initialised Meta model, following the same
        structure, forward etc.
        """
        super().__init__()
        self.input_modules = input_modules
        self.fusion_modules = fusion_modules
        self.output_modules = output_modules

    def forward(
        self, inputs: Dict[str, torch.Tensor]
    ) -> Dict[str, Dict[str, torch.Tensor]]:
        return run_meta_forward(
            input_modules=self.input_modules,
            fusion_modules=self.fusion_modules,
            output_modules=self.output_modules,
            fusion_to_output_mapping={},
            inputs=inputs,
        )


def attach_store_forward_hook(
    module: nn.Module,
    cache: Dict[str, CachedTensor],
    layer_path: str,
    layer_cache_target: Literal["input", "output"],
) -> Callable[[], None]:
    def hook(module: nn.Module, args: Tuple[torch.Tensor, ...], output: torch.Tensor):

        if layer_cache_target == "input":
            cache[layer_path] = CachedTensor(
                tensor=args[0],
                shape=args[0].shape,
                layer_path=layer_path,
            )
        else:
            cache[layer_path] = CachedTensor(
                tensor=output,
                shape=output.shape,
                layer_path=layer_path,
            )

    handle = module.register_forward_hook(hook)

    def remove_hook():
        handle.remove()

    return remove_hook


def attach_tensor_broker_module_injection(
    target_module: nn.Module,
    tensor_broker_module: nn.Module,
    tensor_cache: Dict[str, CachedTensor],
    tensor_cache_key: str,
) -> Callable[[], None]:
    def hook(
        module: nn.Module,
        args: Tuple[torch.Tensor, ...],
        kwargs: Dict[str, torch.Tensor],
    ) -> Tuple[Tuple[torch.Tensor, ...], Dict[str, torch.Tensor]]:
        if len(args) > 0:
            input_tensor = args[0]
        else:
            x = kwargs.get("x")
            input_key = kwargs.get("input")
            if x is not None:
                input_tensor = x
            elif input_key is not None:
                input_tensor = input_key
            else:
                raise ValueError("No input tensor found in args or kwargs.")

        cached_tensor = tensor_cache[tensor_cache_key].tensor
        tensor_broker_out = tensor_broker_module(input_tensor, cached_tensor)

        if len(args) > 0:
            new_args = (tensor_broker_out,) + args[1:]
            return new_args, kwargs
        else:
            new_kwargs = kwargs.copy()
            new_kwargs["x" if "x" in kwargs else "input"] = tensor_broker_out
            return args, new_kwargs

    handle = target_module.register_forward_pre_hook(hook, with_kwargs=True)

    def remove_hook():
        handle.remove()

    return remove_hook


def get_tensor_broker(
    input_objects: al_input_objects_as_dict,
    output_objects: al_output_objects_as_dict,
    input_modules: al_input_modules,
    fusion_modules: al_fusion_modules,
    output_modules: al_output_modules,
    fusion_to_output_mapping: Dict[str, Literal["computed", "pass-through"]],
    input_configs: Sequence[InputConfig],
    fusion_configs: Sequence[FusionConfig],
    output_configs: Sequence[OutputConfig],
    device: str,
) -> nn.ModuleDict:

    tensor_broker_modules = nn.ModuleDict()
    all_configs = list(input_configs) + list(fusion_configs) + list(output_configs)

    any_tensor_broker = any(
        config.tensor_broker_config is not None for config in all_configs
    )
    if not any_tensor_broker:
        return tensor_broker_modules

    module_storage = ModuleStorage(
        input_modules=input_modules,
        fusion_modules=fusion_modules,
        output_modules=output_modules,
    )
    module_storage = module_storage.to(device=device)

    example_batch = prepare_example_test_batch(
        input_objects=input_objects,
        output_objects=output_objects,
        model=module_storage,
        device=device,
    )

    output_shapes, input_shapes = estimate_tensor_shapes(
        module_storage=module_storage,
        fusion_to_output_mapping=fusion_to_output_mapping,
        example_input_data=example_batch.inputs,
    )

    tensor_cache: Dict[str, CachedTensor] = {}
    all_named_modules: dict[str, nn.Module] = dict(module_storage.named_modules())

    have_been_cached_mapping: dict[str, tuple[str, str]] = {}
    for config in all_configs:
        if not config.tensor_broker_config:
            continue

        for tmc in config.tensor_broker_config.message_configs:
            layer_path = tmc.layer_path
            layer_cache_target = tmc.layer_cache_target

            if tmc.cache_tensor:
                msg = (
                    f"When attempting to cache '{tmc.name}', "
                    f"layer path '{layer_path}' was not found in module storage. "
                    f"please check the layer path in the tensor broker config."
                )
                module = fuzzy_dict_lookup(
                    d=all_named_modules,
                    key=layer_path,
                    custom_prefix_message=msg,
                )

                attach_store_forward_hook(
                    module=module,
                    cache=tensor_cache,
                    layer_path=layer_path,
                    layer_cache_target=layer_cache_target,
                )
                have_been_cached_mapping[tmc.name] = (layer_path, layer_cache_target)

    have_been_used_from_cache = set()
    for config in all_configs:
        if not config.tensor_broker_config:
            continue

        for tmc in config.tensor_broker_config.message_configs:
            to_path = tmc.layer_path
            to_name = tmc.name

            if tmc.use_from_cache:
                for from_name in tmc.use_from_cache:

                    from_path, cache_target = have_been_cached_mapping[from_name]
                    message_name = f"{to_name}: {from_path}>>>{to_path}"
                    # . is not allowed in layer names in Torch
                    message_name = message_name.replace(".", "--")

                    if cache_target == "output":
                        from_shape_no_batch = output_shapes[from_path]
                    else:
                        from_shape_no_batch = input_shapes[from_path]

                    msg = (
                        f"When setting up message '{message_name}', "
                        f"the destination path "
                        f"'{to_path}' was not found as a module in the model."
                    )
                    to_shape_no_batch = fuzzy_dict_lookup(
                        d=input_shapes,
                        key=to_path,
                        custom_prefix_message=msg,
                    )

                    projection_layer, projected_shape = get_projection_layer(
                        from_shape_no_batch=from_shape_no_batch,
                        to_shape_no_batch=to_shape_no_batch,
                        cache_fusion_type=tmc.cache_fusion_type,
                        projection_type=tmc.projection_type,
                    )
                    have_been_used_from_cache.add(from_name)

                    fusion_layer = get_fusion_layer_wrapper(
                        projected_shape=projected_shape,
                        target_shape=to_shape_no_batch,
                        cache_fusion_type=tmc.cache_fusion_type,
                        projection_layer=projection_layer,
                        device=device,
                    )

                    tensor_broker_modules[message_name] = fusion_layer

                    attach_tensor_broker_module_injection(
                        target_module=all_named_modules[to_path],
                        tensor_broker_module=fusion_layer,
                        tensor_cache=tensor_cache,
                        tensor_cache_key=from_path,
                    )

    tensor_broker_modules = tensor_broker_modules.to(device=device)
    return tensor_broker_modules


def fuzzy_dict_lookup(
    d: Dict[str, Any],
    key: str,
    num_suggestions: int = 3,
    custom_prefix_message: Optional[str] = None,
) -> Any:
    if key in d:
        return d[key]

    close_matches = get_close_matches(key, d.keys(), n=num_suggestions, cutoff=0.6)

    error_parts = []

    if custom_prefix_message:
        error_parts.append(f"{custom_prefix_message}")

    error_parts.append(f"Key not found: '{key}'")

    if close_matches:
        suggestions = "\n".join(f"  - {match}" for match in close_matches)
        error_parts.append(f"Did you mean one of these?\n{suggestions}")
    else:
        error_parts.append("No close matches found.")

    sample_keys = list(d.keys())[:5]
    if sample_keys:
        examples = "\n".join(f"  - {k}" for k in sample_keys)
        error_parts.append(f"Example keys from the dictionary:\n{examples}")
        if len(d) > 5:
            error_parts.append(f"  ... ({len(d) - 5} more keys)")

    error_message = "\n".join(error_parts)
    logger.error(error_message)

    raise KeyError(f"Key '{key}' not found. See logs for details.")



import json
from pathlib import Path
from typing import Optional

import numpy as np

from eir.experiment_io.io_utils import load_dataclass
from eir.experiment_io.output_object_io_modules.output_io_utils import (
    load_output_config_from_yaml,
)
from eir.setup.output_setup_modules.array_output_setup import (
    ArrayNormalizationStats,
    ComputedArrayOutputInfo,
    DataDimensions,
    DiffusionConfig,
    set_up_array_output,
)


def load_array_output_object(serialized_output_folder: Path) -> ComputedArrayOutputInfo:
    config_path = serialized_output_folder / "output_config.yaml"
    normalization_stats_path = serialized_output_folder / "normalization_stats.json"
    dtype_path = serialized_output_folder / "dtype.json"
    data_dimensions_path = serialized_output_folder / "data_dimensions.json"
    diffusion_path = serialized_output_folder / "diffusion.json"

    output_config = load_output_config_from_yaml(output_config_path=config_path)

    normalization_stats: Optional[ArrayNormalizationStats] = None
    if normalization_stats_path.exists():
        normalization_stats = load_dataclass(
            cls=ArrayNormalizationStats,
            file_path=normalization_stats_path,
        )

    dtype_str = json.loads(dtype_path.read_text())
    dtype = np.dtype(dtype_str)

    data_dimensions = load_dataclass(
        cls=DataDimensions,
        file_path=data_dimensions_path,
    )

    diffusion_config = None
    if diffusion_path.exists():
        diffusion_config = load_dataclass(
            cls=DiffusionConfig,
            file_path=diffusion_path,
        )

    loaded_object = set_up_array_output(
        output_config=output_config,
        normalization_stats=normalization_stats,
        data_dimensions=data_dimensions,
        dtype=dtype,
        diffusion_config=diffusion_config,
    )

    return loaded_object



import re
from functools import partial


def _split_tokenizer(x):  # noqa: F821
    # type: (str) -> list[str]
    return x.split()


def _spacy_tokenize(x, spacy):
    return [tok.text for tok in spacy.tokenizer(x)]


_patterns = [
    r"\'",
    r"\"",
    r"\.",
    r"<br \/>",
    r",",
    r"\(",
    r"\)",
    r"\!",
    r"\?",
    r"\;",
    r"\:",
    r"\s+",
]

_replacements = [
    " '  ",
    "",
    " . ",
    " ",
    " , ",
    " ( ",
    " ) ",
    " ! ",
    " ? ",
    " ",
    " ",
    " ",
]

_patterns_dict = list((re.compile(p), r) for p, r in zip(_patterns, _replacements))


def _basic_english_normalize(line):
    r"""
    Basic normalization for a line of text.
    Normalization includes
    - lowercasing
    - complete some basic text normalization for English words as follows:
        add spaces before and after '\''
        remove '\"',
        add spaces before and after '.'
        replace '<br \/>'with single space
        add spaces before and after ','
        add spaces before and after '('
        add spaces before and after ')'
        add spaces before and after '!'
        add spaces before and after '?'
        replace ';' with single space
        replace ':' with single space
        replace multiple spaces with single space

    Returns a list of tokens after splitting on whitespace.
    """

    line = line.lower()
    for pattern_re, replaced_str in _patterns_dict:
        line = pattern_re.sub(replaced_str, line)
    return line.split()


def get_tokenizer(tokenizer, language="en"):
    r"""
    Generate tokenizer function for a string sentence.

    Args:
        tokenizer: the name of tokenizer function. If None, it returns split()
            function, which splits the string sentence by space.
            If basic_english, it returns _basic_english_normalize() function,
            which normalize the string first and split by space. If a callable
            function, it will return the function. If a tokenizer library
            (e.g. spacy, moses, toktok, revtok, subword), it returns the
            corresponding library.
        language: Default en

    Examples:
        >>> tokenizer = get_tokenizer("basic_english")
        >>> tokens = tokenizer("You can now install TorchText using pip!")
        >>> tokens
        >>> ['you', 'can', 'now', 'install', 'torchtext', 'using', 'pip', '!']

    """

    # default tokenizer is string.split(), added as a module function for serialization
    if tokenizer is None:
        return _split_tokenizer

    if tokenizer == "basic_english":
        if language != "en":
            raise ValueError("Basic normalization is only available for Enlish(en)")
        return _basic_english_normalize

    # simply return if a function is passed
    if callable(tokenizer):
        return tokenizer

    if tokenizer == "spacy":
        try:
            import spacy

            try:
                spacy = spacy.load(language)
            except IOError:
                # Model shortcuts no longer work in spaCy 3.0+, try using fullnames
                # List is from
                # https://github.com/explosion/spaCy/blob/
                # b903de3fcb56df2f7247e5b6cfa6b66f4ff02b62/spacy/errors.py#L789
                OLD_MODEL_SHORTCUTS = (
                    spacy.errors.OLD_MODEL_SHORTCUTS
                    if hasattr(spacy.errors, "OLD_MODEL_SHORTCUTS")
                    else {}
                )
                if language not in OLD_MODEL_SHORTCUTS:
                    raise
                import warnings

                warnings.warn(
                    f'Spacy model "{language}" could not be loaded,'
                    f' trying "{OLD_MODEL_SHORTCUTS[language]}" instead'
                )
                spacy = spacy.load(OLD_MODEL_SHORTCUTS[language])
            return partial(_spacy_tokenize, spacy=spacy)
        except ImportError:
            print(
                "Please install SpaCy. "
                "See the docs at https://spacy.io for more information."
            )
            raise
        except AttributeError:
            print(
                "Please install SpaCy and the SpaCy {} tokenizer. "
                "See the docs at https://spacy.io for more "
                "information.".format(language)
            )
            raise
    elif tokenizer == "moses":
        try:
            from sacremoses import MosesTokenizer

            moses_tokenizer = MosesTokenizer()
            return moses_tokenizer.tokenize
        except ImportError:
            print(
                "Please install SacreMoses. "
                "See the docs at https://github.com/alvations/sacremoses "
                "for more information."
            )
            raise
    elif tokenizer == "toktok":
        try:
            from nltk.tokenize.toktok import ToktokTokenizer

            toktok = ToktokTokenizer()
            return toktok.tokenize
        except ImportError:
            print(
                "Please install NLTK. "
                "See the docs at https://nltk.org  for more information."
            )
            raise
    elif tokenizer == "revtok":
        try:
            import revtok

            return revtok.tokenize
        except ImportError:
            print("Please install revtok.")
            raise
    elif tokenizer == "subword":
        try:
            import revtok

            return partial(revtok.tokenize, decap=True)
        except ImportError:
            print("Please install revtok.")
            raise
    raise ValueError(
        "Requested tokenizer {}, valid choices are a "
        "callable that takes a single string as input, "
        '"revtok" for the revtok reversible tokenizer, '
        '"subword" for the revtok caps-aware tokenizer, '
        '"spacy" for the SpaCy English tokenizer, or '
        '"moses" for the NLTK port of the Moses tokenization '
        "script.".format(tokenizer)
    )



import itertools

from eir.models.input.array.models_cnn import (
    ConvParamSuggestion,
    conv_output_formula,
    solve_for_padding,
)


def calc_conv_params_for_dimension(
    input_size: int,
    target_size: int,
    min_threshold: float,
    max_kernel_size: int = 7,
    max_stride: int = 4,
    max_dilation: int = 3,
    stride_to_kernel_ratio: float = 1.0,
) -> list[ConvParamSuggestion]:
    if input_size < target_size:
        raise NotImplementedError(
            "When using CNN based projection in tensor broker, "
            "target size currently cannot be larger than input size. "
            "This might be happening if you are sending a tensor message with a "
            "smaller spatial size to a larger one. For this, kindly use a different "
            "projection layer (e.g. 'grouped_linear')."
        )

    valid_params = []

    for kernel_size, stride, dilation in itertools.product(
        range(1, max_kernel_size + 1),
        range(1, max_stride + 1),
        range(1, max_dilation + 1),
    ):
        if stride > kernel_size * stride_to_kernel_ratio:
            continue

        padding = solve_for_padding(
            input_size=input_size,
            target_size=target_size,
            dilation=dilation,
            stride=stride,
            kernel_size=kernel_size,
        )

        if padding is None:
            continue

        output_size = conv_output_formula(
            input_size=input_size,
            padding=padding,
            dilation=dilation,
            kernel_size=kernel_size,
            stride=stride,
        )

        size_ratio = output_size / target_size

        if size_ratio >= min_threshold:
            valid_params.append(
                ConvParamSuggestion(
                    kernel_size=kernel_size,
                    stride=stride,
                    dilation=dilation,
                    padding=padding,
                    target_size=output_size,
                )
            )

    if not valid_params:
        raise ValueError(
            f"No valid convolutional parameters found to transform "
            f"{input_size} to {target_size} with min_threshold {min_threshold}"
        )

    return valid_params


def choose_best_params(
    params: list[ConvParamSuggestion], target_size: int
) -> ConvParamSuggestion:
    """
    Slightly confusing here that the output size of the param suggestion is
    named target_size, but it is actually the output size of the convolution.
    """
    return min(
        params,
        key=lambda p: (
            abs(p.target_size - target_size),
            p.kernel_size + p.stride + p.dilation + p.padding,
        ),
    )


def get_conv_params_for_dimension(
    input_size: int,
    target_size: int,
    min_threshold: float = 1.0,
    max_kernel_size: int = 33,
    max_stride: int = 32,
    max_dilation: int = 4,
) -> ConvParamSuggestion:
    valid_params = calc_conv_params_for_dimension(
        input_size=input_size,
        target_size=target_size,
        min_threshold=min_threshold,
        max_kernel_size=max_kernel_size,
        max_stride=max_stride,
        max_dilation=max_dilation,
    )
    best_solution = choose_best_params(
        params=valid_params,
        target_size=target_size,
    )

    return best_solution



import json
from copy import copy
from dataclasses import asdict
from pathlib import Path
from typing import TYPE_CHECKING, Any, Union

import numpy as np
from aislib.misc_utils import ensure_path_exists

from eir.experiment_io.input_object_io_modules.array_input_io import (
    load_array_input_object,
)
from eir.experiment_io.input_object_io_modules.bytes_input_io import (
    load_bytes_input_object,
)
from eir.experiment_io.input_object_io_modules.image_input_io import (
    load_image_input_object,
)
from eir.experiment_io.input_object_io_modules.omics_input_io import (
    load_omics_input_object,
)
from eir.experiment_io.input_object_io_modules.sequence_input_io import (
    load_sequence_input_object,
)
from eir.experiment_io.io_utils import (
    check_version,
    dump_config_to_yaml,
    save_dataclass,
)
from eir.setup import schemas
from eir.setup.input_setup_modules.setup_array import ComputedArrayInputInfo
from eir.setup.input_setup_modules.setup_bytes import ComputedBytesInputInfo
from eir.setup.input_setup_modules.setup_image import ComputedImageInputInfo
from eir.setup.input_setup_modules.setup_omics import ComputedOmicsInputInfo
from eir.setup.input_setup_modules.setup_sequence import (
    ComputedSequenceInputInfo,
    extract_tokenizer_object_from_function,
)
from eir.train_utils.utils import get_logger, get_run_folder

if TYPE_CHECKING:
    from eir.setup.input_setup import (
        al_input_objects_as_dict,
        al_serializable_input_classes,
        al_serializable_input_objects,
    )


logger = get_logger(name=__name__)


def load_serialized_input_object(
    input_config: schemas.InputConfig,
    input_class: "al_serializable_input_classes",
    *args,
    output_folder: Union[None, str] = None,
    run_folder: Union[None, Path] = None,
    custom_input_name: Union[str, None] = None,
    **kwargs,
) -> "al_serializable_input_objects":
    assert output_folder or run_folder
    if not run_folder:
        assert output_folder is not None
        run_folder = get_run_folder(output_folder=output_folder)

    check_version(run_folder=run_folder)
    input_name = input_config.input_info.input_name
    input_type = input_config.input_info.input_type

    if custom_input_name:
        input_name = custom_input_name

    serialized_input_folder_path = get_input_serialization_path(
        run_folder=run_folder,
        input_name=input_name,
        input_type=input_type,
    )

    assert serialized_input_folder_path.exists(), serialized_input_folder_path
    serialized_input_config_object = _read_serialized_input_object(
        input_class=input_class,
        serialized_input_folder=serialized_input_folder_path,
    )

    assert isinstance(serialized_input_config_object, input_class)

    train_input_info_kwargs = serialized_input_config_object.__dict__
    assert "input_config" in train_input_info_kwargs.keys()

    loaded_input_info_kwargs = copy(train_input_info_kwargs)

    _check_current_and_loaded_input_config_compatibility(
        current_input_config=input_config,
        loaded_input_config=serialized_input_config_object.input_config,
        serialized_input_config_path=serialized_input_folder_path,
    )
    loaded_input_info_kwargs["input_config"] = input_config

    loaded_input_object = input_class(**loaded_input_info_kwargs)

    return loaded_input_object


def get_input_serialization_path(
    run_folder: Path,
    input_type: str,
    input_name: str,
) -> Path:

    base_path = run_folder / "serializations" / f"{input_type}_input_serializations"
    match input_type:
        case "image" | "sequence" | "bytes" | "array" | "omics":
            path = base_path / f"{input_name}/"
        case _:
            raise ValueError(f"Invalid input type: {input_type}")

    return path


def _check_current_and_loaded_input_config_compatibility(
    current_input_config: schemas.InputConfig,
    loaded_input_config: schemas.InputConfig,
    serialized_input_config_path: Path,
) -> None:
    fieldnames = current_input_config.__dict__.keys()
    assert set(fieldnames) == set(loaded_input_config.__dict__.keys())

    should_be_same = ("model_config",)

    for key in should_be_same:
        current_value = getattr(current_input_config, key)
        loaded_value = getattr(loaded_input_config, key)

        if _should_skip_warning(
            key=key,
            current_value=current_value,
            loaded_value=loaded_value,
        ):
            continue

        if current_value != loaded_value:
            logger.warning(
                "Expected '%s' to be the same in current input configuration"
                "\n'%s'\n and "
                "loaded input configuration \n'%s'\n "
                "(loaded from '%s'). If you are loading"
                " a pretrained EIR module, this can be expected if you are changing "
                "parameters that are expected to be agnostic across runs (e.g. dropout)"
                ", but in many cases this will cause (a) the model you are trying to "
                "load and (b) the model you are setting up for the current experiment "
                "to diverge, which will most likely lead to a RuntimeError. The "
                "resolution is likely to ensure that the input configurations of "
                "(a) and (b) are exactly the same when it comes to model "
                "configurations, which should ensure that the model architectures "
                "match.",
                key,
                current_value,
                loaded_value,
                serialized_input_config_path,
            )


def _should_skip_warning(key: str, current_value: Any, loaded_value: Any) -> bool:
    if key != "model_config":
        return False

    current_dict = asdict(current_value)
    loaded_dict = asdict(loaded_value)

    differing_keys = [k for k, v in current_dict.items() if loaded_dict.get(k) != v]

    return (
        len(differing_keys) == 1
        and differing_keys[0] == "model_type"
        and "linked" in loaded_dict["model_type"]
    )


def serialize_chosen_input_objects(
    inputs_dict: "al_input_objects_as_dict", run_folder: Path
) -> None:
    targets_to_serialize = {
        "sequence",
        "bytes",
        "image",
        "array",
        "omics",
    }

    for input_name, input_ in inputs_dict.items():
        input_type = input_.input_config.input_info.input_type

        any_match = any(i for i in targets_to_serialize if input_type == i)

        if any_match:
            output_path = get_input_serialization_path(
                run_folder=run_folder,
                input_type=input_type,
                input_name=input_name,
            )

            assert isinstance(
                input_,
                (
                    ComputedImageInputInfo,
                    ComputedSequenceInputInfo,
                    ComputedBytesInputInfo,
                    ComputedArrayInputInfo,
                    ComputedOmicsInputInfo,
                ),
            )

            ensure_path_exists(path=output_path, is_folder=output_path.is_dir())
            _serialize_input_object(
                input_object=input_,
                output_folder=output_path,
            )


def _serialize_input_object(
    input_object: "al_serializable_input_objects",
    output_folder: Path,
) -> None:

    input_config = input_object.input_config
    config_path = output_folder / "input_config.yaml"

    match input_object:
        case ComputedImageInputInfo():
            dump_config_to_yaml(config=input_config, output_path=config_path)

            save_dataclass(
                obj=input_object.normalization_stats,
                file_path=output_folder / "normalization_stats.json",
            )

            num_channels_obj = {"num_channels": input_object.num_channels}
            with open(output_folder / "num_channels.json", "w") as f:
                json.dump(num_channels_obj, f)

        case ComputedSequenceInputInfo():
            dump_config_to_yaml(config=input_config, output_path=config_path)

            index_to_string = input_object.vocab.itos
            with open(output_folder / "vocab_ordered.txt", "w") as f:
                for token in index_to_string:
                    f.write(f"{token}\n")

            string_to_index = input_object.vocab.stoi
            with open(output_folder / "vocab.json", "w") as f:
                json.dump(string_to_index, f)

            input_type_info = input_config.input_type_info
            assert isinstance(input_type_info, schemas.SequenceInputDataConfig)
            tokenizer = input_type_info.tokenizer
            if tokenizer == "bpe":
                tokenizer_callable = input_object.tokenizer
                assert tokenizer_callable is not None
                tokenizer_object = extract_tokenizer_object_from_function(
                    tokenizer_callable=tokenizer_callable
                )
                tokenizer_object.save(str(output_folder / "bpe_tokenizer.json"))

            computed_max_length = input_object.computed_max_length
            with open(output_folder / "computed_max_length.json", "w") as f:
                json.dump(computed_max_length, f)

        case ComputedBytesInputInfo():
            dump_config_to_yaml(config=input_config, output_path=config_path)

            with open(output_folder / "vocab.json", "w") as f:
                json.dump(input_object.vocab, f)

            computed_max_length = input_object.computed_max_length
            with open(output_folder / "computed_max_length.json", "w") as f:
                json.dump(computed_max_length, f)

        case ComputedArrayInputInfo():
            dump_config_to_yaml(config=input_config, output_path=config_path)

            if input_object.normalization_stats is not None:
                save_dataclass(
                    obj=input_object.normalization_stats,
                    file_path=output_folder / "normalization_stats.json",
                )

            dtype_str = str(input_object.dtype)
            with open(output_folder / "dtype.json", "w") as f:
                json.dump(dtype_str, f)

            save_dataclass(
                obj=input_object.data_dimensions,
                file_path=output_folder / "data_dimensions.json",
            )

        case ComputedOmicsInputInfo():
            dump_config_to_yaml(config=input_config, output_path=config_path)

            input_type_info = input_config.input_type_info
            assert isinstance(input_type_info, schemas.OmicsInputDataConfig)

            subset_indices_file = input_type_info.subset_snps_file
            if subset_indices_file:
                subset_indices_file_src = Path(subset_indices_file)
                subset_indices_file_dst = output_folder / "subset_snps_file.txt"
                subset_indices_file_dst.write_text(subset_indices_file_src.read_text())

            snp_file = input_type_info.snp_file
            if snp_file:
                snp_file_src = Path(snp_file)
                snp_file_dst = output_folder / "snps.bim"
                snp_file_dst.write_text(snp_file_src.read_text())

            save_dataclass(
                obj=input_object.data_dimensions,
                file_path=output_folder / "data_dimensions.json",
            )

            subset_indices = input_object.subset_indices
            if subset_indices is not None:
                np.save(file=output_folder / "subset_indices.npy", arr=subset_indices)


def _read_serialized_input_object(
    input_class: "al_serializable_input_classes",
    serialized_input_folder: Path,
) -> "al_serializable_input_objects":
    loaded_object: "al_serializable_input_objects"
    if input_class is ComputedImageInputInfo:
        loaded_object = load_image_input_object(
            serialized_input_folder=serialized_input_folder
        )

    elif input_class is ComputedSequenceInputInfo:
        loaded_object = load_sequence_input_object(
            serialized_input_folder=serialized_input_folder
        )

    elif input_class is ComputedBytesInputInfo:
        loaded_object = load_bytes_input_object(
            serialized_input_folder=serialized_input_folder
        )

    elif input_class is ComputedArrayInputInfo:
        loaded_object = load_array_input_object(
            serialized_input_folder=serialized_input_folder
        )

    elif input_class is ComputedOmicsInputInfo:
        loaded_object = load_omics_input_object(
            serialized_input_folder=serialized_input_folder
        )

    else:
        raise ValueError(f"Invalid input class: {input_class}")

    return loaded_object



import copy
import logging

from termcolor import colored
from tqdm import tqdm


class TQDMLoggingHandler(logging.Handler):
    def __init__(self, level=logging.NOTSET):
        super().__init__(level)

    def emit(self, record):
        try:
            msg = self.format(record)
            tqdm.write(msg)
            self.flush()
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            self.handleError(record)


class ColoredFormatter(logging.Formatter):
    def format(self, record):
        record_copy = copy.copy(record)

        if record_copy.levelno == logging.DEBUG:
            record_copy.levelname = colored(record_copy.levelname, "blue")
        elif record_copy.levelno == logging.INFO:
            record_copy.levelname = colored(record_copy.levelname, "green")
        elif record_copy.levelno == logging.WARNING:
            record_copy.levelname = colored(record_copy.levelname, "yellow")
        elif record_copy.levelno == logging.ERROR:
            record_copy.levelname = colored(record_copy.levelname, "red")

        return super().format(record_copy)


def get_logger(name: str, tqdm_compatible: bool = False) -> logging.Logger:
    """
    Creates a logger with a debug level and a custom format.

    tqdm_compatible: Overwrite default stream.write in favor of tqdm.write
    to avoid breaking progress bar.
    """
    logger_ = logging.getLogger(name)
    logger_.setLevel(logging.DEBUG)

    handler: logging.Handler | TQDMLoggingHandler
    if tqdm_compatible:
        handler = TQDMLoggingHandler()
    else:
        handler = logging.StreamHandler()

    formatter = ColoredFormatter(
        fmt="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
        datefmt="%H:%M:%S",
    )
    handler.setFormatter(formatter)

    logger_.addHandler(handler)
    return logger_



import json
from copy import copy
from pathlib import Path
from typing import TYPE_CHECKING, Sequence

from aislib.misc_utils import ensure_path_exists

from eir.experiment_io.io_utils import (
    check_version,
    dump_config_to_yaml,
    save_dataclass,
)
from eir.experiment_io.output_object_io_modules.array_output_io import (
    load_array_output_object,
)
from eir.experiment_io.output_object_io_modules.image_output_io import (
    load_image_output_object,
)
from eir.experiment_io.output_object_io_modules.sequence_output_io import (
    load_sequence_output_object,
)
from eir.experiment_io.output_object_io_modules.survival_output_io import (
    load_survival_output_object,
)
from eir.experiment_io.output_object_io_modules.tabular_output_io import (
    load_tabular_output_object,
)
from eir.setup import schemas
from eir.setup.input_setup_modules.setup_sequence import (
    extract_tokenizer_object_from_function,
)
from eir.setup.output_setup_modules.array_output_setup import ComputedArrayOutputInfo
from eir.setup.output_setup_modules.image_output_setup import ComputedImageOutputInfo
from eir.setup.output_setup_modules.sequence_output_setup import (
    ComputedSequenceOutputInfo,
)
from eir.setup.output_setup_modules.survival_output_setup import (
    ComputedSurvivalOutputInfo,
)
from eir.setup.output_setup_modules.tabular_output_setup import (
    ComputedTabularOutputInfo,
)

if TYPE_CHECKING:
    from eir.setup.output_setup import (
        al_output_classes,
        al_output_objects,
        al_output_objects_as_dict,
    )


def load_all_serialized_output_objects(
    output_configs: Sequence[schemas.OutputConfig],
    run_folder: Path,
) -> "al_output_objects_as_dict":
    output_objects_as_dict = {}

    output_class: "al_output_classes"
    for output_config in output_configs:
        output_type = output_config.output_info.output_type
        match output_type:
            case "tabular":
                output_class = ComputedTabularOutputInfo
            case "sequence":
                output_class = ComputedSequenceOutputInfo
            case "image":
                output_class = ComputedImageOutputInfo
            case "array":
                output_class = ComputedArrayOutputInfo
            case "survival":
                output_class = ComputedSurvivalOutputInfo
            case _:
                raise ValueError(f"Invalid output type: {output_type}")

        output_object = load_serialized_output_object(
            output_config=output_config,
            output_class=output_class,
            run_folder=run_folder,
        )
        output_objects_as_dict[output_config.output_info.output_name] = output_object

    return output_objects_as_dict


def load_serialized_output_object(
    output_config: schemas.OutputConfig,
    output_class: "al_output_classes",
    run_folder: Path,
) -> "al_output_objects":
    check_version(run_folder=run_folder)

    output_name = output_config.output_info.output_name
    output_type = output_config.output_info.output_type

    serialized_output_folder = get_output_serialization_path(
        run_folder=run_folder,
        output_type=output_type,
        output_name=output_name,
    )

    assert serialized_output_folder.exists(), serialized_output_folder
    serialized_output_config_object = _read_serialized_output_object(
        output_class=output_class,
        serialized_output_folder=serialized_output_folder,
        run_folder=run_folder,
    )
    assert isinstance(serialized_output_config_object, output_class)

    output_info_kwargs = serialized_output_config_object.__dict__
    assert "output_config" in output_info_kwargs

    loaded_kwargs = copy(output_info_kwargs)
    loaded_kwargs["output_config"] = output_config

    loaded_output_object = output_class(**loaded_kwargs)

    return loaded_output_object


def get_output_serialization_path(
    run_folder: Path,
    output_type: str,
    output_name: str,
) -> Path:

    base_path = run_folder / "serializations" / f"{output_type}_output_serializations"
    match output_type:
        case "image" | "sequence" | "array" | "tabular" | "survival":
            path = base_path / f"{output_name}/"
        case _:
            raise ValueError(f"Invalid output type: {output_type}")

    return path


def serialize_output_objects(
    output_objects: "al_output_objects_as_dict",
    run_folder: Path,
) -> None:

    for output_name, output_object in output_objects.items():
        output_type = output_object.output_config.output_info.output_type
        output_path = get_output_serialization_path(
            run_folder=run_folder,
            output_type=output_type,
            output_name=output_name,
        )
        ensure_path_exists(path=output_path, is_folder=output_path.is_dir())

        serialize_output_object(
            output_object=output_object,
            output_folder=output_path,
        )


def serialize_output_object(
    output_object: "al_output_objects",
    output_folder: Path,
) -> None:
    output_config = output_object.output_config
    config_path = output_folder / "output_config.yaml"

    match output_object:
        case ComputedTabularOutputInfo() | ComputedSurvivalOutputInfo():
            dump_config_to_yaml(config=output_config, output_path=config_path)
        case ComputedSequenceOutputInfo():
            dump_config_to_yaml(config=output_config, output_path=config_path)

            index_to_string = output_object.vocab.itos
            with open(output_folder / "vocab_ordered.txt", "w") as f:
                for token in index_to_string:
                    f.write(f"{token}\n")

            string_to_index = output_object.vocab.stoi
            with open(output_folder / "vocab.json", "w") as f:
                json.dump(string_to_index, f)

            output_type_info = output_config.output_type_info
            assert isinstance(output_type_info, schemas.SequenceOutputTypeConfig)
            tokenizer = output_type_info.tokenizer
            if tokenizer == "bpe":
                tokenizer_callable = output_object.tokenizer
                assert tokenizer_callable is not None
                tokenizer_object = extract_tokenizer_object_from_function(
                    tokenizer_callable=tokenizer_callable
                )
                tokenizer_object.save(str(output_folder / "bpe_tokenizer.json"))

            computed_max_length = output_object.computed_max_length
            with open(output_folder / "computed_max_length.json", "w") as f:
                json.dump(computed_max_length, f)

            embedding_dim = output_object.embedding_dim
            with open(output_folder / "embedding_dim.json", "w") as f:
                json.dump(embedding_dim, f)

        case ComputedImageOutputInfo():
            dump_config_to_yaml(config=output_config, output_path=config_path)

            save_dataclass(
                obj=output_object.normalization_stats,
                file_path=output_folder / "normalization_stats.json",
            )

            num_channels_obj = {"num_channels": output_object.num_channels}
            with open(output_folder / "num_channels.json", "w") as f:
                json.dump(num_channels_obj, f)

            if output_object.diffusion_config is not None:
                save_dataclass(
                    obj=output_object.diffusion_config,
                    file_path=output_folder / "diffusion_config.json",
                )

        case ComputedArrayOutputInfo():
            dump_config_to_yaml(config=output_config, output_path=config_path)

            if output_object.normalization_stats is not None:
                save_dataclass(
                    obj=output_object.normalization_stats,
                    file_path=output_folder / "normalization_stats.json",
                )

            dtype_str = str(output_object.dtype)
            with open(output_folder / "dtype.json", "w") as f:
                json.dump(dtype_str, f)

            save_dataclass(
                obj=output_object.data_dimensions,
                file_path=output_folder / "data_dimensions.json",
            )

            if output_object.diffusion_config is not None:
                save_dataclass(
                    obj=output_object.diffusion_config,
                    file_path=output_folder / "diffusion_config.json",
                )

        case _:
            raise ValueError(f"Invalid output object: {output_object}")


def _read_serialized_output_object(
    output_class: "al_output_classes",
    serialized_output_folder: Path,
    run_folder: Path,
) -> "al_output_objects":
    loaded_object: "al_output_objects"
    if output_class is ComputedTabularOutputInfo:
        loaded_object = load_tabular_output_object(
            serialized_output_folder=serialized_output_folder,
            run_folder=run_folder,
        )
    elif output_class is ComputedSurvivalOutputInfo:
        loaded_object = load_survival_output_object(
            serialized_output_folder=serialized_output_folder,
            run_folder=run_folder,
        )
    elif output_class is ComputedSequenceOutputInfo:
        loaded_object = load_sequence_output_object(
            serialized_output_folder=serialized_output_folder,
            run_folder=run_folder,
        )
    elif output_class is ComputedImageOutputInfo:
        loaded_object = load_image_output_object(
            serialized_output_folder=serialized_output_folder,
        )
    elif output_class is ComputedArrayOutputInfo:
        loaded_object = load_array_output_object(
            serialized_output_folder=serialized_output_folder,
        )
    else:
        raise ValueError(f"Invalid output class: {output_class}")

    return loaded_object



from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Literal

from eir.experiment_io.configs_io import load_configs
from eir.experiment_io.io_utils import check_version
from eir.experiment_io.label_transformer_io import load_transformers
from eir.experiment_io.output_object_io import load_all_serialized_output_objects
from eir.setup.config import Configs
from eir.setup.output_setup import al_output_objects_as_dict
from eir.train_utils.metrics import get_default_metrics
from eir.train_utils.step_logic import get_default_hooks
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.train_utils.metrics import al_metric_record_dict
    from eir.train_utils.step_logic import Hooks


logger = get_logger(name=__name__, tqdm_compatible=True)


@dataclass
class LoadedTrainExperiment:
    configs: Configs
    hooks: "Hooks"
    metrics: "al_metric_record_dict"
    outputs: al_output_objects_as_dict


def load_serialized_train_experiment(
    run_folder: Path,
    device: str,
    source_folder: Literal["configs", "configs_stripped"] = "configs_stripped",
) -> LoadedTrainExperiment:
    check_version(run_folder=run_folder)

    if source_folder == "configs":
        configs_folder = run_folder / "configs"
    elif source_folder == "configs_stripped":
        configs_folder = run_folder / "serializations" / "configs_stripped"
    else:
        raise ValueError(
            f"source_folder must be one of 'configs' or 'configs_stripped', "
            f"not {source_folder}"
        )

    configs_loaded = load_configs(configs_root_folder=configs_folder)

    configs_loaded.global_config.be.device = device

    hooks = get_default_hooks(configs=configs_loaded)

    transformers = load_transformers(run_folder=run_folder)
    gc = configs_loaded.global_config
    metrics = get_default_metrics(
        target_transformers=transformers,
        cat_metrics=gc.metrics.cat_metrics,
        con_metrics=gc.metrics.con_metrics,
        cat_averaging_metrics=gc.metrics.cat_averaging_metrics,
        con_averaging_metrics=gc.metrics.con_averaging_metrics,
        output_configs=configs_loaded.output_configs,
    )

    output_objects_as_dict = load_all_serialized_output_objects(
        output_configs=configs_loaded.output_configs,
        run_folder=run_folder,
    )

    loaded_experiment = LoadedTrainExperiment(
        configs=configs_loaded,
        hooks=hooks,
        metrics=metrics,
        outputs=output_objects_as_dict,
    )

    return loaded_experiment


def get_version_file(run_folder: Path) -> Path:
    return run_folder / "meta/eir_version.txt"



import dataclasses
from typing import Callable, Dict

import torch
from torch import nn

from eir.models.fusion.fusion_default import default_fuse_features

al_identity_features = Callable[
    [Dict[str, torch.Tensor]], torch.Tensor | Dict[str, torch.Tensor]
]


@dataclasses.dataclass
class IdentityConfig:
    pass


class IdentityFusionModel(nn.Module):
    def __init__(
        self,
        model_config: IdentityConfig,
        fusion_in_dim: int,
        fusion_callable: al_identity_features = default_fuse_features,
        **kwargs,
    ):
        super().__init__()

        self.model_config = model_config
        self.fusion_in_dim = fusion_in_dim
        self.fusion_callable = fusion_callable

        self._init_weights()

    def _init_weights(self):
        pass

    @property
    def num_out_features(self) -> int:
        return self.fusion_in_dim

    def forward(
        self, inputs: Dict[str, torch.Tensor]
    ) -> torch.Tensor | Dict[str, torch.Tensor]:
        fused_features = self.fusion_callable(inputs)

        return fused_features



import json
from copy import deepcopy
from pathlib import Path

from eir.experiment_io.input_object_io import get_input_serialization_path
from eir.experiment_io.input_object_io_modules.sequence_input_io import (
    load_sequence_input_object,
)
from eir.experiment_io.output_object_io_modules.output_io_utils import (
    load_output_config_from_yaml,
)
from eir.setup.output_setup_modules.sequence_output_setup import (
    ComputedSequenceOutputInfo,
    set_up_sequence_output,
)
from eir.setup.schemas import SequenceOutputTypeConfig


def load_sequence_output_object(
    serialized_output_folder: Path,
    run_folder: Path,
) -> ComputedSequenceOutputInfo:
    config_path = serialized_output_folder / "output_config.yaml"
    vocab_json_path = _get_vocab_path(serialized_output_folder=serialized_output_folder)
    computed_max_length_path = serialized_output_folder / "computed_max_length.json"

    output_config = load_output_config_from_yaml(output_config_path=config_path)
    computed_max_length = json.loads(computed_max_length_path.read_text())

    output_name = output_config.output_info.output_name
    matching_input_path = get_input_serialization_path(
        run_folder=run_folder,
        input_name=output_name,
        input_type="sequence",
    )
    matching_input_object = load_sequence_input_object(
        serialized_input_folder=matching_input_path,
    )

    output_config_modified = deepcopy(output_config)
    output_type_info_modified = deepcopy(output_config.output_type_info)
    assert isinstance(output_type_info_modified, SequenceOutputTypeConfig)

    output_type_info_modified.vocab_file = str(vocab_json_path)
    output_type_info_modified.max_length = computed_max_length

    output_config_modified.output_type_info = output_type_info_modified

    loaded_object = set_up_sequence_output(
        output_config=output_config_modified,
        input_objects={output_name: matching_input_object},
    )

    return loaded_object


def _get_vocab_path(serialized_output_folder: Path) -> Path:
    vocab_json_path = serialized_output_folder / "vocab.json"
    bpe_tokenizer_path = serialized_output_folder / "bpe_tokenizer.json"
    if bpe_tokenizer_path.exists():
        return bpe_tokenizer_path
    return vocab_json_path



from pathlib import Path
from typing import Optional

import numpy as np
import torch

from eir.data_load.data_preparation_modules.common import _load_deeplake_sample
from eir.data_load.data_source_modules import deeplake_ops
from eir.setup.input_setup_modules.setup_array import ArrayNormalizationStats


def array_load_wrapper(
    data_pointer: Path | int,
    input_source: str,
    deeplake_inner_key: Optional[str] = None,
) -> np.ndarray:
    if deeplake_ops.is_deeplake_dataset(data_source=input_source):
        assert deeplake_inner_key is not None
        assert isinstance(data_pointer, int)
        array_data = _load_deeplake_sample(
            data_pointer=data_pointer,
            input_source=input_source,
            inner_key=deeplake_inner_key,
        )
    else:
        assert isinstance(data_pointer, (str, Path)), data_pointer
        array_data = np.load(str(data_pointer))

    assert isinstance(array_data, np.ndarray)
    return array_data


def prepare_array_data(
    array_data: np.ndarray,
    normalization_stats: Optional[ArrayNormalizationStats],
) -> torch.Tensor:
    """Enforce 3 dimensions for now."""

    tensor = torch.from_numpy(array_data).float()

    match len(tensor.shape):
        case 1:
            tensor = tensor.unsqueeze(dim=0).unsqueeze(dim=0)
        case 2:
            tensor = tensor.unsqueeze(dim=0)
        case 3:
            tensor = tensor
        case _:
            raise ValueError(
                f"Array has {len(tensor.shape)} dimensions, currently only "
                f"1, 2, or 3 are supported."
            )

    tensor_no_nan = fill_nans_from_stats(
        array=tensor,
        normalization_stats=normalization_stats,
    )
    tensor_normalized = normalize_array(
        array=tensor_no_nan,
        normalization_stats=normalization_stats,
    )

    return tensor_normalized


def fill_nans_from_stats(
    array: torch.Tensor,
    normalization_stats: Optional[ArrayNormalizationStats],
) -> torch.Tensor:
    array = array.to(device="cpu")
    nan_mask = torch.isnan(array)
    if not nan_mask.any():
        return array

    if normalization_stats is None:
        return torch.where(nan_mask, torch.zeros_like(array), array)

    match normalization_stats.type:
        case "channel":
            return torch.where(
                nan_mask, normalization_stats.means.expand_as(array), array
            )
        case "element":
            return torch.where(nan_mask, normalization_stats.means, array)
        case _:
            raise ValueError(f"Unknown normalization type: {normalization_stats.type}")


def normalize_array(
    array: torch.Tensor,
    normalization_stats: Optional[ArrayNormalizationStats],
    epsilon: float = 1e-10,
) -> torch.Tensor:
    if normalization_stats is None:
        return array

    means = normalization_stats.means
    stds = normalization_stats.stds
    array = (array.to(device="cpu") - means) / (stds + epsilon)

    return array


def un_normalize_array(
    array: torch.Tensor, normalization_stats: Optional[ArrayNormalizationStats]
) -> torch.Tensor:
    if normalization_stats is None:
        return array

    means = normalization_stats.means
    stds = normalization_stats.stds
    array = array.to(device="cpu") * stds + means

    return array



from dataclasses import dataclass
from typing import Optional

import torch
import torch.nn.functional as F
from einops import rearrange
from torch import Tensor, nn

from eir.models.layers.norm_layers import RMSNorm


class LinearAttention(nn.Module):
    """
    From https://github.com/lucidrains
    """

    def __init__(
        self,
        embed_dim: int,
        heads: int = 4,
        dim_head: int = 32,
    ):
        super().__init__()
        self.scale = dim_head**-0.5
        self.heads = heads
        hidden_dim = dim_head * heads

        self.to_qkv = nn.Conv2d(
            in_channels=embed_dim,
            out_channels=hidden_dim * 3,
            kernel_size=1,
            bias=False,
        )

        self.to_out = nn.Sequential(
            nn.Conv2d(in_channels=hidden_dim, out_channels=embed_dim, kernel_size=1),
            RMSNorm(dim=embed_dim, normalize_dim=1),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        b, c, h, w = x.shape

        qkv = self.to_qkv(x).chunk(3, dim=1)
        q, k, v = map(
            lambda t: rearrange(t, "b (h c) x y -> b h c (x y)", h=self.heads), qkv
        )

        q = q.softmax(dim=-2)
        k = k.softmax(dim=-1)

        q = q * self.scale

        context = torch.einsum("b h d n, b h e n -> b h d e", k, v)

        out = torch.einsum("b h d e, b h d n -> b h e n", context, q)
        out = rearrange(out, "b h c (x y) -> b (h c) x y", h=self.heads, x=h, y=w)

        return self.to_out(out)


class SwiGLU(nn.Module):
    def __init__(
        self,
        in_features: int,
        hidden_features: int,
        out_features: int,
        bias: bool = True,
    ):
        super().__init__()
        self.w1 = nn.Linear(
            in_features=in_features,
            out_features=hidden_features,
            bias=bias,
        )
        self.w2 = nn.Linear(
            in_features=in_features,
            out_features=hidden_features,
            bias=bias,
        )
        self.w3 = nn.Linear(
            in_features=hidden_features,
            out_features=out_features,
            bias=bias,
        )

    def _init_weights(self):
        torch.nn.init.normal_(self.w1.weight, mean=0.0, std=0.02)
        torch.nn.init.normal_(self.w1.weight, mean=0.0, std=0.02)
        torch.nn.init.normal_(self.w1.weight, mean=0.0, std=0.02)

    def forward(self, x: Tensor) -> Tensor:
        hidden = self.w1(x)
        gate = self.w2(x)
        hidden = F.silu(hidden) * gate
        return self.w3(hidden)


@dataclass
class TransformerBlockConfig:
    rope_theta: int = 10000


class TransformerBlock(nn.Module):
    def __init__(
        self,
        d_model: int,
        n_head: int,
        dim_feedforward: int,
        dropout: float = 0.1,
        norm_first: bool = True,
        block_config: Optional[TransformerBlockConfig] = None,
    ):
        super().__init__()

        msg = "d_model ({d_model}) must be divisible by n_head ({n_head})"
        assert d_model % n_head == 0, msg

        self.d_model = d_model
        self.n_head = n_head
        self.head_dim = d_model // n_head
        self.norm_first = norm_first
        self.config = block_config or TransformerBlockConfig()

        self.q_proj = nn.Linear(d_model, d_model, bias=False)
        self.k_proj = nn.Linear(d_model, d_model, bias=False)
        self.v_proj = nn.Linear(d_model, d_model, bias=False)
        self.out_proj = nn.Linear(d_model, d_model, bias=False)

        self.norm1 = nn.RMSNorm(normalized_shape=d_model)
        self.norm2 = nn.RMSNorm(normalized_shape=d_model)

        self.ffn = SwiGLU(
            in_features=d_model,
            hidden_features=dim_feedforward,
            out_features=d_model,
            bias=False,
        )

        self.dropout = nn.Dropout(dropout)

        self._init_rope()

    def _init_rope(self):
        dim = self.head_dim
        theta = self.config.rope_theta
        freqs = 1.0 / (theta ** (torch.arange(0, dim, 2).float() / dim))
        self.register_buffer("rope_freqs", freqs)

    def _apply_rope(self, x: Tensor, seq_len: int) -> Tensor:
        position = torch.arange(seq_len, device=x.device)
        freqs = torch.outer(position, self.rope_freqs)

        # Create rotation matrices
        cos = torch.cos(freqs).view(seq_len, 1, -1)
        sin = torch.sin(freqs).view(seq_len, 1, -1)

        # Split input into even and odd dimensions
        x_even = x[..., ::2]
        x_odd = x[..., 1::2]

        # Apply rotation
        rotated = torch.stack(
            [x_even * cos - x_odd * sin, x_odd * cos + x_even * sin], dim=-1
        ).flatten(-2)

        return rotated

    def _attention(self, x: Tensor, attn_mask: Optional[Tensor] = None) -> Tensor:
        batch_size, seq_len, _ = x.shape

        # Project to queries, keys, values and reshape
        q = self.q_proj(x).view(batch_size, seq_len, self.n_head, self.head_dim)
        k = self.k_proj(x).view(batch_size, seq_len, self.n_head, self.head_dim)
        v = self.v_proj(x).view(batch_size, seq_len, self.n_head, self.head_dim)

        # Apply rotary embeddings
        q = self._apply_rope(x=q, seq_len=seq_len)
        k = self._apply_rope(x=k, seq_len=seq_len)

        # Transpose for attention: [batch_size, num_heads, sequence_length, head_dim]
        q, k, v = [t.transpose(1, 2) for t in (q, k, v)]

        attn_output = F.scaled_dot_product_attention(
            query=q,
            key=k,
            value=v,
            attn_mask=attn_mask,
            dropout_p=self.dropout.p if self.training else 0.0,
        )

        # Reshape and project output
        attn_output = (
            attn_output.transpose(1, 2)
            .contiguous()
            .view(batch_size, seq_len, self.d_model)
        )
        return self.dropout(self.out_proj(attn_output))

    def forward(self, x: Tensor, attn_mask: Optional[Tensor] = None) -> Tensor:
        if self.norm_first:
            x = x + self._attention(self.norm1(x), attn_mask)
            x = x + self.ffn(self.norm2(x))
        else:
            x = self.norm1(x + self._attention(x, attn_mask))
            x = self.norm2(x + self.ffn(x))
        return x


class Transformer(nn.Module):

    def __init__(
        self,
        d_model: int,
        nhead: int,
        num_layers: int,
        dim_feedforward: int,
        dropout: float = 0.1,
        norm_first: bool = True,
    ):
        super().__init__()

        self.layers = nn.ModuleList(
            [
                TransformerBlock(
                    d_model=d_model,
                    n_head=nhead,
                    dim_feedforward=dim_feedforward,
                    dropout=dropout,
                    norm_first=norm_first,
                )
                for _ in range(num_layers)
            ]
        )

    def forward(self, x: Tensor, mask: Optional[Tensor] = None) -> Tensor:
        for layer in self.layers:
            x = layer(x, mask)
        return x



from dataclasses import dataclass
from typing import Literal, Union

SurvivalLossNames = Literal["NegativeLogLikelihood", "CoxPHLoss"]


@dataclass
class SurvivalOutputTypeConfig:
    """
    Basic configuration for survival analysis output.

    :param time_column:
        The name of the column in the label file that contains the time-to-event or
        censoring time.

    :param event_column:
        The name of the column in the label file that indicates whether an event
        occurred (1) or the observation was censored (0).

    :param num_durations:
        The number of discrete time intervals to use in the model. This determines the
        size of the output layer.

    :param loss_function:
        The loss function to use for training the survival model.

    :param max_duration:
        The maximum duration to consider. Times beyond this will be censored at this
        point. If None, use the maximum observed time.

    :param label_parsing_chunk_size:
        Number of rows to process at a time when loading the input_source.
        Useful when RAM is limited.
    """

    time_column: str
    event_column: str
    num_durations: int = 10
    loss_function: SurvivalLossNames = "NegativeLogLikelihood"
    max_duration: Union[None, float] = None
    label_parsing_chunk_size: Union[None, int] = None



import sys
from dataclasses import dataclass
from enum import Enum
from typing import Any, Optional, Type

import numpy as np
import polars as pl
import torch

from eir.utils.logging import get_logger

logger = get_logger(name=__name__)


class ColumnType(Enum):
    NUMERIC = "numeric"
    STRING = "string"
    PATH = "path"
    FIXED_ARRAY = "fixed_array"
    VAR_ARRAY = "var_array"
    OBJECT = "object"


@dataclass
class ColumnInfo:
    name: str
    dtype: ColumnType
    original_dtype: str
    index: int
    array_shape: Optional[tuple[int, ...]] = None
    inner_dtype: Optional[str] = None
    torch_dtype: Optional[torch.dtype] = None


class HybridStorage:
    """
    Storage for mixed-type data using torch.Tensor for numeric data
    and numpy arrays for other data types that cannot be neatly
    represented as tensors.
    """

    def __init__(self):
        self.numeric_data: Optional[torch.Tensor] = None
        self.string_data: Optional[np.ndarray] = None
        self.path_data: Optional[np.ndarray] = None
        self.fixed_array_data: Optional[list[torch.Tensor]] = None
        self.var_array_data: Optional[list[np.ndarray]] = None
        self.object_data: Optional[dict[str, list[object]]] = None

        self.numeric_columns: list[ColumnInfo] = []
        self.string_columns: list[ColumnInfo] = []
        self.path_columns: list[ColumnInfo] = []
        self.fixed_array_columns: list[ColumnInfo] = []
        self.var_array_columns: list[ColumnInfo] = []
        self.object_columns: list[ColumnInfo] = []
        self._num_rows: int = 0

    def from_polars(self, df: pl.DataFrame) -> None:
        self._num_rows = len(df)

        logger.debug("Converting Polars DataFrame to HybridStorage.")

        for idx, col in enumerate(df.columns):
            series = df.get_column(col)
            dtype = series.dtype
            dtype_str = str(dtype)

            is_fixed, inner_type, shape = _is_fixed_array(dtype=dtype)
            if is_fixed:
                self.fixed_array_columns.append(
                    ColumnInfo(
                        name=col,
                        dtype=ColumnType.FIXED_ARRAY,
                        original_dtype=dtype_str,
                        index=idx,
                        array_shape=shape,
                        inner_dtype=inner_type,
                    )
                )
                continue

            is_var, inner_type = _is_var_array(dtype=dtype)
            if is_var:
                self.var_array_columns.append(
                    ColumnInfo(
                        name=col,
                        dtype=ColumnType.VAR_ARRAY,
                        original_dtype=dtype_str,
                        index=idx,
                        inner_dtype=inner_type,
                    )
                )
                continue

            if _is_numeric_dtype(dtype=dtype):
                torch_dtype = (
                    torch.float32 if "float" in dtype_str.lower() else torch.int64
                )
                self.numeric_columns.append(
                    ColumnInfo(
                        name=col,
                        dtype=ColumnType.NUMERIC,
                        original_dtype=dtype_str,
                        index=idx,
                        torch_dtype=torch_dtype,
                    )
                )
            elif _is_path_dtype(name=col, values=series):
                self.path_columns.append(
                    ColumnInfo(
                        name=col,
                        dtype=ColumnType.PATH,
                        original_dtype=dtype_str,
                        index=idx,
                    )
                )
            elif _is_string_dtype(dtype=dtype):
                self.string_columns.append(
                    ColumnInfo(
                        name=col,
                        dtype=ColumnType.STRING,
                        original_dtype=dtype_str,
                        index=idx,
                    )
                )
            else:
                self.object_columns.append(
                    ColumnInfo(
                        name=col,
                        dtype=ColumnType.OBJECT,
                        original_dtype=dtype_str,
                        index=idx,
                    )
                )

        if self.numeric_columns:
            numeric_data = []
            for col_info in self.numeric_columns:
                series = df.get_column(name=col_info.name)
                numeric_data.append(torch.tensor(series.to_numpy()))
            self.numeric_data = torch.stack(numeric_data) if numeric_data else None

        if self.fixed_array_columns:
            fixed_array_data = []
            for col_info in self.fixed_array_columns:
                series = df.get_column(name=col_info.name)
                samples_np = series.to_numpy().astype(col_info.inner_dtype)
                arr = torch.tensor(samples_np)
                fixed_array_data.append(arr)

            self.fixed_array_data = fixed_array_data

        if self.var_array_columns:
            self.var_array_data = []
            for col_info in self.var_array_columns:
                series = df.get_column(name=col_info.name)
                # Store as numpy array for variable length lists
                self.var_array_data.append(series.to_numpy())

        if self.string_columns:
            string_data = []
            for col_info in self.string_columns:
                series = df.get_column(name=col_info.name)
                string_data.append(series.to_numpy())
            self.string_data = np.array(string_data, dtype=np.str_)

        if self.path_columns:
            path_data = []
            for col_info in self.path_columns:
                series = df.get_column(name=col_info.name)
                path_data.append(series.to_numpy())
            self.path_data = np.array(path_data, dtype=np.str_)

        if self.object_columns:
            if self.object_data is None:
                self.object_data = {}

            for col_info in self.object_columns:
                series = df.get_column(name=col_info.name)
                cur_object_data = series.to_list()
                self.object_data[col_info.name] = cur_object_data

    def get_ids(self) -> list[str]:
        id_col_info = next(col for col in self.string_columns if col.name == "ID")
        assert self.string_data is not None
        return self.string_data[self.string_columns.index(id_col_info)].tolist()

    def __len__(self) -> int:
        return self._num_rows

    def __repr__(self) -> str:
        lines = [f"HybridStorage(rows={self._num_rows})"]

        def format_columns(title: str, columns: list[ColumnInfo]) -> list[str]:
            if not columns:
                return []
            result = [f"\n{title} Columns:"]
            for col in columns:
                base = f"  {col.name} ({col.original_dtype})"
                if col.array_shape:
                    base += f" shape={col.array_shape}"
                if col.inner_dtype:
                    base += f" inner_type={col.inner_dtype}"
                result.append(base)
            return result

        sections = [
            ("Numeric", self.numeric_columns),
            ("Fixed Array", self.fixed_array_columns),
            ("Variable Array", self.var_array_columns),
            ("String", self.string_columns),
            ("Path", self.path_columns),
            ("Object", self.object_columns),
        ]

        for title, cols in sections:
            lines.extend(format_columns(title, cols))

        return "\n".join(lines)

    def __str__(self) -> str:
        counts = {
            "numeric": len(self.numeric_columns),
            "fixed_array": len(self.fixed_array_columns),
            "var_array": len(self.var_array_columns),
            "string": len(self.string_columns),
            "path": len(self.path_columns),
            "object": len(self.object_columns),
        }

        type_counts = [f"{k}={v}" for k, v in counts.items() if v > 0]

        return f"HybridStorage(rows={self._num_rows}, " + ", ".join(type_counts) + ")"

    def check_data(self) -> None:
        """
        Checks for:
            1. Non-empty storage
            2. Valid ID column
            3. At least one non-null value per row
        """
        if self._num_rows == 0:
            raise ValueError(
                "Expected to have at least one sample, but got empty storage. "
                "Possibly there is a mismatch between input IDs and target IDs."
            )

        id_col_info = next(
            (col for col in self.string_columns if col.name == "ID"), None
        )
        if not id_col_info:
            raise ValueError("Could not find required 'ID' column in string columns")

        assert self.string_data is not None
        id_idx = self.string_columns.index(id_col_info)
        id_data = self.string_data[id_idx]

        null_id_mask = np.array([not id_ for id_ in id_data])
        if null_id_mask.any():
            null_indices = np.where(null_id_mask)[0]
            raise ValueError(
                f"Expected all observations to have a sample ID associated "
                f"with them, but got rows with null IDs at indices: "
                f"{null_indices.tolist()}"
            )

        masks = []

        if self.numeric_data is not None:
            numeric_mask = torch.isnan(self.numeric_data).all(dim=0).numpy()
            masks.append(numeric_mask)

        if self.string_data is not None:
            string_indices = [i for i in range(len(self.string_columns)) if i != id_idx]
            if string_indices:
                string_mask = np.all(
                    [
                        np.array([not bool(x) for x in self.string_data[i]])
                        for i in string_indices
                    ],
                    axis=0,
                )
                masks.append(string_mask)

        if self.fixed_array_data is not None:
            fixed_array_masks = []
            for col_data in self.fixed_array_data:
                nan_mask = torch.isnan(input=col_data)
                other_dims = tuple(range(1, len(nan_mask.shape)))
                fixed_array_mask = torch.all(input=nan_mask, dim=other_dims).numpy()
                fixed_array_masks.append(fixed_array_mask)

            masks.append(np.all(fixed_array_masks, axis=0))

        if self.var_array_data:
            var_masks: list[np.ndarray] = []
            for arr_data in self.var_array_data:
                cur_mask: list[bool] = []
                for arr in arr_data:
                    cur_mask.append(arr is None or len(arr) == 0)
                var_masks.append(np.array(cur_mask))

            var_array_mask = np.all(var_masks, axis=0)
            masks.append(var_array_mask)

        if self.path_data is not None:
            path_mask = np.all(check_empty_str_arr(data=self.path_data), axis=0)
            masks.append(path_mask)

        if self.object_data:
            object_mask = np.all(
                [
                    np.array([x is None for x in col_data])
                    for col_data in self.object_data.values()
                ],
                axis=0,
            )
            masks.append(object_mask)

        if masks:
            all_null_mask = np.all(masks, axis=0)
            null_rows = np.where(all_null_mask)[0]

            if len(null_rows) > 0:
                bad_id = id_data[null_rows[0]]
                raise ValueError(
                    f"Expected all observations to have at least one input value "
                    f"associated with them, but got empty inputs for ID: {bad_id}"
                )

    def validate_storage(self, name: str = "") -> int:
        self.check_data()

        memory_bytes = get_total_memory(
            numeric_data=self.numeric_data,
            string_data=self.string_data,
            path_data=self.path_data,
            fixed_array_data=self.fixed_array_data,
            var_array_data=self.var_array_data,
            object_data=self.object_data,
        )

        memory_value, memory_unit = format_memory_size(bytes_size=memory_bytes)

        logger.info(
            f"{name} is using {memory_value:.4f}{memory_unit} of memory "
            f"holding {self._num_rows} samples."
        )

        return memory_bytes

    def get_row(self, idx: int) -> dict[str, Any]:
        result: dict[str, Any] = {}

        if self.numeric_data is not None:
            # the isnan check there is to guard against the case where
            # if we have nan, casting directly to long will corrupt / convert it to 0
            # in torch
            for i, col_info in enumerate(self.numeric_columns):
                value_tensor = self.numeric_data[i, idx]
                if torch.isnan(value_tensor):
                    value = np.nan
                else:
                    value = value_tensor.to(dtype=col_info.torch_dtype).item()

                result[col_info.name] = value

        if self.fixed_array_data is not None:
            for i, col_info in enumerate(self.fixed_array_columns):
                result[col_info.name] = self.fixed_array_data[i][idx].numpy()

        if self.var_array_data is not None:
            for i, col_info in enumerate(self.var_array_columns):
                result[col_info.name] = self.var_array_data[i][idx]

        if self.string_data is not None:
            for i, col_info in enumerate(self.string_columns):
                result[col_info.name] = self.string_data[i, idx].item()

        if self.path_data is not None:
            for i, col_info in enumerate(self.path_columns):
                result[col_info.name] = self.path_data[i, idx].item()

        if self.object_data is not None:
            for i, col_info in enumerate(self.object_columns):
                result[col_info.name] = self.object_data[col_info.name][idx]

        return result


def check_empty_str_arr(data: np.ndarray) -> bool:
    str_arr = data.astype(str)
    return (str_arr == "") | (str_arr == "nan")


def _recursive_find_primitive_dtype(dtype: pl.DataType) -> Type[pl.DataType]:
    if hasattr(dtype, "inner"):
        return _recursive_find_primitive_dtype(dtype=dtype.inner)
    return type(dtype)


def _is_numeric_dtype(dtype: pl.DataType) -> bool:
    return dtype.is_numeric()


def _is_fixed_array(
    dtype: pl.DataType,
) -> tuple[bool, Optional[str], Optional[tuple]]:
    if not dtype == pl.Array:
        return False, None, None

    assert isinstance(dtype, pl.Array)

    inner_dtype = dtype.inner
    assert isinstance(inner_dtype, pl.DataType)

    inner_type = _recursive_find_primitive_dtype(dtype=inner_dtype)
    inner_dtype_str = polars_dtype_to_str_dtype(polars_dtype=inner_type)
    shape = dtype.shape

    return True, inner_dtype_str, shape


def _is_var_array(dtype: pl.DataType) -> tuple[bool, Optional[str]]:
    if not dtype == pl.List:
        return False, None

    assert isinstance(dtype, pl.List)
    inner_type = dtype.inner
    return True, str(inner_type)


def _is_path_dtype(name: str, values: pl.Series) -> bool:
    if not isinstance(values[0], str):
        return False
    return any(
        p in str(values[0]).lower()
        for p in ["/var/", "/tmp/", ".csv", ".txt", ".json", ".npy"]
    )


def _is_string_dtype(dtype: pl.DataType) -> bool:
    return dtype == pl.String


def check_two_storages(
    input_storage: HybridStorage,
    target_storage: Optional[HybridStorage],
) -> None:
    if target_storage is None or target_storage._num_rows == 0:
        return

    masks = []

    if target_storage.numeric_data is not None:
        numeric_mask = torch.isnan(target_storage.numeric_data).all(dim=0).numpy()
        masks.append(numeric_mask)

    if target_storage.string_data is not None:
        id_idx = next(
            i for i, col in enumerate(target_storage.string_columns) if col.name == "ID"
        )
        string_indices = [
            i for i in range(len(target_storage.string_columns)) if i != id_idx
        ]
        if string_indices:
            string_mask = np.all(
                [
                    np.array([not bool(x) for x in target_storage.string_data[i]])
                    for i in string_indices
                ],
                axis=0,
            )
            masks.append(string_mask)

    if target_storage.fixed_array_data is not None:
        fixed_array_masks = []
        for col_data in target_storage.fixed_array_data:
            nan_mask = torch.isnan(input=col_data)
            other_dims = tuple(range(1, len(nan_mask.shape)))
            fixed_array_mask = torch.all(input=nan_mask, dim=other_dims).numpy()
            fixed_array_masks.append(fixed_array_mask)

        masks.append(np.all(fixed_array_masks, axis=0))

    if target_storage.var_array_data:
        var_masks: list[np.ndarray] = []
        for arr_data in target_storage.var_array_data:
            cur_mask: list[bool] = []
            for arr in arr_data:
                cur_mask.append(arr is None or len(arr) == 0)
            var_masks.append(np.array(cur_mask))

        var_array_mask = np.all(var_masks, axis=0)
        masks.append(var_array_mask)

    if target_storage.object_data:
        object_mask = np.all(
            [
                np.array([x is None for x in col_data])
                for col_data in target_storage.object_data.values()
            ],
            axis=0,
        )
        masks.append(object_mask)

    if masks:
        all_null_mask = np.all(masks, axis=0)
        null_rows = np.where(all_null_mask)[0]
        if len(null_rows) > 0:
            bad_id = target_storage.get_ids()[null_rows[0]]
            raise ValueError(
                f"Expected all observations to have at least one target label "
                f"associated with them, but got empty targets for ID: {bad_id}"
            )

    input_ids = set(input_storage.get_ids())
    target_ids = set(target_storage.get_ids())

    if input_ids != target_ids:
        missing_in_input = target_ids - input_ids
        missing_in_target = input_ids - target_ids
        msg = []
        if missing_in_input:
            msg.append(f"IDs in target but not in input: {missing_in_input}")
        if missing_in_target:
            msg.append(f"IDs in input but not in target: {missing_in_target}")
        raise ValueError(
            "ID mismatch between input and target storage. " + " ".join(msg)
        )


def is_null_value(value: Any) -> bool:
    if isinstance(value, (np.ndarray, torch.Tensor)):
        return False
    if isinstance(value, str):
        return value in ("", "None", "null", "nan", "__NULL__")
    if isinstance(value, (int, float)):
        return np.isnan(value) if isinstance(value, float) else False
    return value is None


def polars_dtype_to_str_dtype(polars_dtype: Type[pl.DataType]) -> str:
    dtype_map = {
        pl.Float64: "float64",
        pl.Float32: "float32",
        pl.Int64: "int64",
        pl.Int32: "int32",
        pl.Int16: "int16",
        pl.Int8: "int8",
        pl.UInt64: "uint64",
        pl.UInt32: "uint32",
        pl.UInt16: "uint16",
        pl.UInt8: "uint8",
        pl.Boolean: "bool",
        pl.Date: "date",
        pl.Datetime: "datetime",
        pl.Utf8: "str",
    }

    return dtype_map[polars_dtype]


def get_numeric_memory(data: Optional[torch.Tensor]) -> int:
    if data is None:
        return 0
    return data.element_size() * data.nelement()


def get_fixed_array_memory(data: Optional[list[torch.Tensor]]) -> int:
    if data is None:
        return 0
    return sum(t.element_size() * t.nelement() for t in data)


def get_array_memory(
    data: Optional[np.ndarray],
) -> int:
    memory_usage = 0
    if data is not None:
        memory_usage += data.nbytes
    return memory_usage


def get_path_memory(data: Optional[np.ndarray]) -> int:
    if data is None:
        return 0
    return data.nbytes


def get_var_array_memory(data: Optional[list[np.ndarray]]) -> int:
    if data is None:
        return 0
    return sum(arr.nbytes for arr in data)


def get_object_memory(data: Optional[dict[str, list[object]]]) -> int:
    if data is None:
        return 0
    return sum(sum(sys.getsizeof(obj) for obj in values) for values in data.values())


def format_memory_size(bytes_size: int) -> tuple[float, str]:
    mb_size = bytes_size / (1024**2)
    if mb_size < 1024:
        return mb_size, "MB"
    return mb_size / 1024, "GB"


def get_total_memory(
    numeric_data: Optional[torch.Tensor] = None,
    string_data: Optional[np.ndarray] = None,
    path_data: Optional[np.ndarray] = None,
    fixed_array_data: Optional[list[torch.Tensor]] = None,
    var_array_data: Optional[list[np.ndarray]] = None,
    object_data: Optional[dict[str, list[object]]] = None,
) -> int:
    return sum(
        [
            get_numeric_memory(data=numeric_data),
            get_fixed_array_memory(data=fixed_array_data),
            get_array_memory(data=string_data),
            get_path_memory(data=path_data),
            get_var_array_memory(data=var_array_data),
            get_object_memory(data=object_data),
        ]
    )



from dataclasses import dataclass
from typing import Optional

from torchvision.transforms import Compose

from eir.models.output.array.array_output_modules import ArrayOutputModuleConfig
from eir.setup.input_setup_modules.common import DataDimensions
from eir.setup.input_setup_modules.setup_image import (
    ImageNormalizationStats,
    get_image_normalization_values,
    get_image_transforms,
    get_num_channels_wrapper,
)
from eir.setup.schemas import ImageOutputTypeConfig, OutputConfig
from eir.train_utils.step_modules.diffusion import (
    DiffusionConfig,
    initialize_diffusion_config,
)
from eir.utils.logging import get_logger

logger = get_logger(name=__name__, tqdm_compatible=True)


@dataclass
class ComputedImageOutputInfo:
    output_config: OutputConfig
    base_transforms: Compose
    all_transforms: Compose
    data_dimensions: DataDimensions
    num_channels: int
    normalization_stats: Optional[ImageNormalizationStats] = None
    diffusion_config: Optional[DiffusionConfig] = None


def set_up_image_output(
    output_config: OutputConfig,
    normalization_stats: Optional[ImageNormalizationStats] = None,
    diffusion_config: Optional[DiffusionConfig] = None,
    *args,
    **kwargs,
) -> ComputedImageOutputInfo:

    output_info = output_config.output_info
    output_type_info = output_config.output_type_info
    assert isinstance(output_type_info, ImageOutputTypeConfig)

    image_mode = output_type_info.mode
    num_channels = output_type_info.num_channels

    num_channels = get_num_channels_wrapper(
        image_mode=image_mode,
        num_channels=num_channels,
        source=output_info.output_source,
        deeplake_inner_key=output_info.output_inner_key,
    )

    data_dimensions = DataDimensions(
        channels=num_channels,
        height=output_type_info.size[0],
        width=output_type_info.size[-1],
    )

    output_type_info = output_config.output_type_info
    assert isinstance(output_type_info, ImageOutputTypeConfig)

    oti = output_type_info
    model_config = output_config.model_config
    assert isinstance(model_config, ArrayOutputModuleConfig)

    if normalization_stats is None:
        normalization_stats = get_image_normalization_values(
            source=output_config.output_info.output_source,
            inner_key=output_config.output_info.output_inner_key,
            model_config=model_config,
            mean_normalization_values=oti.mean_normalization_values,
            stds_normalization_values=oti.stds_normalization_values,
            adaptive_normalization_max_samples=oti.adaptive_normalization_max_samples,
            data_dimensions=data_dimensions,
            image_mode=oti.mode,
        )

    base_transforms, all_transforms = get_image_transforms(
        target_size=(data_dimensions.height, data_dimensions.width),
        normalization_stats=normalization_stats,
        auto_augment=False,
        resize_approach=oti.resize_approach,
    )

    if diffusion_config is None:
        if output_type_info.loss == "diffusion":
            time_steps = output_type_info.diffusion_time_steps
            if time_steps is None:
                raise ValueError(
                    "Diffusion loss requires specifying the number of time steps."
                    "Please set `diffusion_time_steps` in the output config."
                )
            diffusion_config = initialize_diffusion_config(time_steps=time_steps)

    image_output_object = ComputedImageOutputInfo(
        output_config=output_config,
        data_dimensions=data_dimensions,
        base_transforms=base_transforms,
        all_transforms=all_transforms,
        num_channels=num_channels,
        normalization_stats=normalization_stats,
        diffusion_config=diffusion_config,
    )

    return image_output_object



import argparse
import types
from collections import Counter
from copy import copy
from dataclasses import dataclass, field
from typing import (
    Any,
    Dict,
    Generator,
    Iterable,
    List,
    Literal,
    Mapping,
    Sequence,
    Tuple,
    Type,
    Union,
)

import configargparse
import yaml

from eir.models.fusion.fusion_identity import IdentityConfig
from eir.models.fusion.fusion_mgmoe import MGMoEModelConfig
from eir.models.layers.mlp_layers import ResidualMLPConfig
from eir.setup import schemas
from eir.setup.config_setup_modules.config_setup_utils import (
    get_yaml_iterator_with_injections,
    validate_keys_against_dataclass,
)
from eir.setup.config_setup_modules.input_config_initialization import init_input_config
from eir.setup.config_setup_modules.output_config_initialization import (
    init_output_config,
)
from eir.setup.config_setup_modules.output_config_setup_sequence import (
    get_configs_object_with_seq_output_configs,
)
from eir.setup.config_validation import validate_train_configs
from eir.setup.schema_modules.latent_analysis_schemas import LatentSamplingConfig
from eir.setup.schema_modules.output_schemas_tabular import TabularOutputTypeConfig
from eir.setup.schemas import (
    AttributionAnalysisConfig,
    BasicExperimentConfig,
    EvaluationCheckpointConfig,
    GlobalConfig,
    GlobalModelConfig,
    LRScheduleConfig,
    OptimizationConfig,
    SupervisedMetricsConfig,
    TrainingControlConfig,
    VisualizationLoggingConfig,
)
from eir.setup.tensor_broker_setup import set_up_tensor_broker_config
from eir.train_utils.utils import configure_global_eir_logging
from eir.utils.logging import get_logger

al_input_types = Union[
    schemas.OmicsInputDataConfig,
    schemas.TabularInputDataConfig,
    schemas.SequenceInputDataConfig,
    schemas.ByteInputDataConfig,
]

logger = get_logger(name=__name__)


def get_configs():
    main_cl_args, extra_cl_args = get_main_cl_args()

    output_folder, log_level = get_output_folder_and_log_level_from_cl_args(
        main_cl_args=main_cl_args,
        extra_cl_args=extra_cl_args,
    )

    configure_global_eir_logging(output_folder=output_folder, log_level=log_level)

    configs = generate_aggregated_config(
        cl_args=main_cl_args,
        extra_cl_args_overload=extra_cl_args,
    )

    configs_with_seq_outputs = get_configs_object_with_seq_output_configs(
        configs=configs,
    )
    validate_train_configs(configs=configs)

    return configs_with_seq_outputs


def get_main_cl_args() -> Tuple[argparse.Namespace, List[str]]:
    parser_ = get_main_parser()
    cl_args, extra_cl_args = parser_.parse_known_args()

    return cl_args, extra_cl_args


def get_output_folder_and_log_level_from_cl_args(
    main_cl_args: argparse.Namespace,
    extra_cl_args: List[str],
) -> Tuple[str, str]:
    global_configs = main_cl_args.global_configs

    output_folder = None
    for config in global_configs:
        with open(config, "r") as f:
            config_dict = yaml.safe_load(f)

            output_folder = config_dict.get("basic_experiment", {}).get("output_folder")
            log_level = config_dict.get("visualization_logging", {}).get(
                "log_level", "info"
            )

            if output_folder and log_level:
                break

    for arg in extra_cl_args:
        if "output_folder=" in arg:
            output_folder = arg.split("=")[1]
            break

    if output_folder is None:
        raise ValueError("Output folder not found in global configs.")

    return output_folder, log_level


def get_main_parser(
    global_nargs: Literal["+", "*"] = "+", output_nargs: Literal["+", "*"] = "+"
) -> configargparse.ArgumentParser:
    parser_ = configargparse.ArgumentParser(
        config_file_parser_class=configargparse.YAMLConfigFileParser
    )

    global_required = True if global_nargs == "+" else False
    output_required = True if output_nargs == "+" else False

    parser_.add_argument(
        "--global_configs",
        nargs=global_nargs,
        type=str,
        required=global_required,
        help="Global .yaml configurations for the experiment.",
    )

    parser_.add_argument(
        "--input_configs",
        type=str,
        nargs="*",
        required=False,
        default=[],
        help="Input feature extraction .yaml configurations. "
        "Each configuration represents one input.",
    )

    parser_.add_argument(
        "--fusion_configs",
        type=str,
        nargs="*",
        required=False,
        default=[],
        help="Fusion .yaml configurations.",
    )

    parser_.add_argument(
        "--output_configs",
        type=str,
        nargs=output_nargs,
        required=output_required,
        help="Output .yaml configurations.",
    )

    return parser_


def _recursive_search(
    dict_: Mapping, target: Any, path=None
) -> Generator[Tuple[str, Any], None, None]:
    if not path:
        path = []

    for key, value in dict_.items():
        local_path = copy(path)
        local_path.append(key)
        if isinstance(value, Mapping):
            yield from _recursive_search(dict_=value, target=target, path=local_path)
        else:
            if value == target:
                yield local_path, value


@dataclass
class Configs:
    global_config: schemas.GlobalConfig
    input_configs: Sequence[schemas.InputConfig]
    fusion_config: schemas.FusionConfig
    output_configs: Sequence[schemas.OutputConfig]

    gc: schemas.GlobalConfig = field(init=False, repr=False)

    def __post_init__(self):
        self.gc = self.global_config


def generate_aggregated_config(
    cl_args: Union[argparse.Namespace, types.SimpleNamespace],
    extra_cl_args_overload: Union[List[str], None] = None,
    strict: bool = True,
) -> Configs:
    global_config_iter = get_yaml_iterator_with_injections(
        yaml_config_files=cl_args.global_configs, extra_cl_args=extra_cl_args_overload
    )
    global_config = get_global_config(global_configs=global_config_iter)

    inputs_config_iter = get_yaml_iterator_with_injections(
        yaml_config_files=cl_args.input_configs,
        extra_cl_args=extra_cl_args_overload,
    )
    input_configs = get_input_configs(input_configs=inputs_config_iter)

    fusion_config_iter = get_yaml_iterator_with_injections(
        yaml_config_files=cl_args.fusion_configs,
        extra_cl_args=extra_cl_args_overload,
    )
    fusion_config = load_fusion_configs(fusion_configs=fusion_config_iter)

    output_config_iter = get_yaml_iterator_with_injections(
        yaml_config_files=cl_args.output_configs,
        extra_cl_args=extra_cl_args_overload,
    )
    output_configs = load_output_configs(output_configs=output_config_iter)

    if strict:
        _check_input_and_output_config_names(
            input_configs=input_configs, output_configs=output_configs
        )

    aggregated_configs = Configs(
        global_config=global_config,
        input_configs=input_configs,
        fusion_config=fusion_config,
        output_configs=output_configs,
    )

    return aggregated_configs


def get_global_config(global_configs: Iterable[dict]) -> GlobalConfig:
    combined_config = combine_dicts(dicts=global_configs)

    basic_experiment_config = BasicExperimentConfig(
        **combined_config.get("basic_experiment", {})
    )
    model_config = GlobalModelConfig(**combined_config.get("model", {}))
    optimization_config = OptimizationConfig(**combined_config.get("optimization", {}))
    lr_schedule_config = LRScheduleConfig(**combined_config.get("lr_schedule", {}))
    training_control_config = TrainingControlConfig(
        **combined_config.get("training_control", {})
    )
    evaluation_checkpoint_config = EvaluationCheckpointConfig(
        **combined_config.get("evaluation_checkpoint", {})
    )
    attribution_analysis_config = AttributionAnalysisConfig(
        **combined_config.get("attribution_analysis", {})
    )
    metrics_config = SupervisedMetricsConfig(**combined_config.get("metrics", {}))
    visualization_logging_config = VisualizationLoggingConfig(
        **combined_config.get("visualization_logging", {})
    )

    latent_sampling = None
    if "latent_sampling" in combined_config and isinstance(
        combined_config["latent_sampling"], dict
    ):
        latent_sampling = LatentSamplingConfig(**combined_config["latent_sampling"])

    global_config = GlobalConfig(
        basic_experiment=basic_experiment_config,
        model=model_config,
        optimization=optimization_config,
        lr_schedule=lr_schedule_config,
        training_control=training_control_config,
        evaluation_checkpoint=evaluation_checkpoint_config,
        attribution_analysis=attribution_analysis_config,
        metrics=metrics_config,
        visualization_logging=visualization_logging_config,
        latent_sampling=latent_sampling,
    )

    return modify_global_config(global_config)


def modify_global_config(global_config: GlobalConfig) -> GlobalConfig:
    gc_copy = copy(global_config)

    if gc_copy.basic_experiment.valid_size > 1.0:
        gc_copy.basic_experiment.valid_size = int(gc_copy.basic_experiment.valid_size)

    return gc_copy


def get_input_configs(
    input_configs: Iterable[dict],
) -> Sequence[schemas.InputConfig]:
    config_objects = []

    for config_dict in input_configs:
        config_object = init_input_config(yaml_config_as_dict=config_dict)
        config_objects.append(config_object)

    _check_input_config_names(input_configs=config_objects)
    return config_objects


def _check_input_config_names(input_configs: Iterable[schemas.InputConfig]) -> None:
    names = []
    for config in input_configs:
        input_name = config.input_info.input_name

        model_name = config.model_config.model_type
        if model_name.startswith("eir-input-sequence-from-linked-output-"):
            logger.info(
                "Skipping input config name check for sequence input config "
                "(name='%s') as it was automatically generated from a sequence output.",
                input_name,
            )
            continue

        names.append(input_name)

    if len(set(names)) != len(names):
        counts = Counter(names)
        raise ValueError(
            f"Found duplicates of input names in input configs: {counts}. "
            f"Please make sure each input source has a unique input_name."
        )


def load_fusion_configs(fusion_configs: Iterable[dict]) -> schemas.FusionConfig:
    combined_config = combine_dicts(dicts=fusion_configs)

    validate_keys_against_dataclass(
        input_dict=combined_config, dataclass_type=schemas.FusionConfig
    )

    combined_config.setdefault("model_type", "mlp-residual")
    combined_config.setdefault("model_config", {})

    fusion_model_type = combined_config["model_type"]

    model_dataclass_config_class: (
        Type[ResidualMLPConfig] | Type[MGMoEModelConfig] | Type[IdentityConfig]
    ) = ResidualMLPConfig
    if fusion_model_type == "mgmoe":
        model_dataclass_config_class = MGMoEModelConfig
    elif fusion_model_type == "linear":
        model_dataclass_config_class = IdentityConfig

    fusion_config_kwargs = combined_config["model_config"]
    fusion_model_config = model_dataclass_config_class(**fusion_config_kwargs)

    tensor_broker_config = set_up_tensor_broker_config(
        tensor_broker_config=combined_config.get("tensor_broker_config", {})
    )

    fusion_config = schemas.FusionConfig(
        model_type=combined_config["model_type"],
        model_config=fusion_model_config,
        tensor_broker_config=tensor_broker_config,
    )

    return fusion_config


def load_output_configs(
    output_configs: Iterable[dict],
) -> Sequence[schemas.OutputConfig]:
    output_config_objects = []

    for config_dict in output_configs:
        config_object = init_output_config(
            yaml_config_as_dict=config_dict,
        )
        output_config_objects.append(config_object)

    _check_output_config_names(output_configs=output_config_objects)
    return output_config_objects


def _check_output_config_names(output_configs: Iterable[schemas.OutputConfig]) -> None:
    names = []
    for config in output_configs:
        names.append(config.output_info.output_name)

    if len(set(names)) != len(names):
        counts = Counter(names)
        raise ValueError(
            f"Found duplicates of input names in input configs: {counts}. "
            f"Please make sure each input source has a unique input_name."
        )


@dataclass
class TabularTargets:
    con_targets: Dict[str, Sequence[str]]
    cat_targets: Dict[str, Sequence[str]]

    def __len__(self):
        all_con = sum(len(i) for i in self.con_targets.values())
        all_cat = sum(len(i) for i in self.cat_targets.values())
        return all_con + all_cat


def get_all_tabular_targets(
    output_configs: Iterable[schemas.OutputConfig],
) -> TabularTargets:
    con_targets = {}
    cat_targets = {}

    for output_config in output_configs:
        if output_config.output_info.output_type != "tabular":
            continue

        output_name = output_config.output_info.output_name
        output_type_info = output_config.output_type_info
        assert isinstance(output_type_info, TabularOutputTypeConfig)
        con_targets[output_name] = output_type_info.target_con_columns
        cat_targets[output_name] = output_type_info.target_cat_columns

    targets = TabularTargets(con_targets=con_targets, cat_targets=cat_targets)
    return targets


def _check_input_and_output_config_names(
    input_configs: Sequence[schemas.InputConfig],
    output_configs: Sequence[schemas.OutputConfig],
    skip_keys: Sequence[str] = ("sequence", "array", "image"),
) -> None:
    """
    We allow for the same name to be used for sequence inputs and outputs,
    as it's used for specifying e.g. different model settings for the input feature
    extractor vs. the output module.
    """
    input_names = set(i.input_info.input_name for i in input_configs)
    output_names = set(
        i.output_info.output_name
        for i in output_configs
        if i.output_info.output_type not in skip_keys
    )

    intersection = output_names.intersection(input_names)
    if len(intersection) > 0:
        raise ValueError(
            "Found common names in input and output configs. Please ensure"
            " that there are no common names between the input and output configs."
            " Input config names: '%s'.\n"
            " Output config names: '%s'.\n",
            " Common names: '%s'.\n",
            input_names,
            output_names,
            intersection,
        )


def combine_dicts(dicts: Iterable[dict]) -> dict:
    combined_dict: dict = {}

    for dict_ in dicts:
        combined_dict = {**combined_dict, **dict_}

    return combined_dict



from dataclasses import dataclass
from typing import Literal, Union

from eir.models.output.tabular.linear import LinearOutputModuleConfig
from eir.models.output.tabular.mlp_residual import ResidualMLPOutputModuleConfig
from eir.models.output.tabular.shared_mlp_residual import (
    SharedResidualMLPOutputModuleConfig,
)


@dataclass
class TabularOutputModuleConfig:
    """
    :param model_init_config:
          Configuration / arguments used to initialise model.

    :param model_type:
         Which type of image model to use."""

    model_init_config: Union[
        ResidualMLPOutputModuleConfig,
        LinearOutputModuleConfig,
        SharedResidualMLPOutputModuleConfig,
    ]
    model_type: Literal[
        "mlp_residual",
        "linear",
        "shared_mlp_residual",
    ] = "mlp_residual"



from typing import Protocol, Type

from eir.models.input.sequence.transformer_models import (
    SequenceOutputTransformerFeatureExtractor,
    TransformerFeatureExtractor,
    TransformerWrapperModel,
)


class SequenceModelClassGetterFunction(Protocol):
    def __call__(
        self, model_type: str
    ) -> (
        Type[TransformerFeatureExtractor]
        | Type[TransformerWrapperModel]
        | Type[SequenceOutputTransformerFeatureExtractor]
    ): ...


def get_sequence_model_class(
    model_type: str,
) -> (
    Type[TransformerFeatureExtractor] | Type[SequenceOutputTransformerFeatureExtractor]
):
    match model_type:
        case "sequence-default":
            return TransformerFeatureExtractor
        case "eir-input-sequence-from-linked-output-default":
            return SequenceOutputTransformerFeatureExtractor
        case _:
            raise ValueError("Invalid sequence model type.")



import inspect
import math
from dataclasses import dataclass
from functools import partial
from typing import (
    Callable,
    Dict,
    Literal,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    Type,
    Union,
)

import torch
from torch import Tensor, nn
from torch.nn.functional import pad
from transformers import PretrainedConfig, PreTrainedModel

from eir.models.layers.attention_layers import Transformer
from eir.models.layers.lcl_layers import _find_lcl_padding_needed
from eir.utils.logging import get_logger

logger = get_logger(name=__name__, tqdm_compatible=True)


@dataclass
class SequenceModelConfig:
    """
    :param model_init_config:
          Configuration / arguments used to initialise model.

    :param model_type:
         Which type of image model to use.

    :param embedding_dim:
        Which dimension to use for the embeddings. If ``None``, will automatically set
        this value based on the number of tokens and attention heads.

    :param position:
        Whether to encode the token position or use learnable position embeddings.

    :param position_dropout:
        Dropout for the positional encoding / embedding.

    :param window_size:
        If set to more than 0, will apply a sliding window of feature
        extraction over the input, meaning the model (e.g. transformer) will only
        see a part of the input at a time. Can be Useful to avoid the O(n²)
        complexity of transformers, as it becomes O(window_size² * n_windows) instead.

    :param pool:
        Whether and how to pool (max / avg) the final feature maps before being
        passed to the final fusion module / predictor. Meaning we pool over the
        sequence (i.e. time) dimension, so the resulting dimensions is embedding_dim
        instead of sequence_length * embedding_dim. If using windowed / conv
        transformers, this becomes embedding_dim * number_of_chunks.

    :param pretrained_model:
          Specify whether the model type is assumed to be pretrained and from the
          Pytorch Image Models repository.

    :param freeze_pretrained_model:
          Whether to freeze the pretrained model weights.

    """

    model_init_config: Union["BasicTransformerFeatureExtractorModelConfig", Dict]

    model_type: Literal["sequence-default"] | str = "sequence-default"

    embedding_dim: int = 64

    position: Literal["encode", "embed"] = "encode"
    position_dropout: float = 0.10
    window_size: int = 0
    pool: Union[Literal["avg"], Literal["max"], None] = None

    pretrained_model: bool = False
    freeze_pretrained_model: bool = False


class TransformerWrapperModel(nn.Module):
    def __init__(
        self,
        feature_extractor: Union["TransformerFeatureExtractor", nn.Module],
        model_config: SequenceModelConfig,
        embedding_dim: int,
        num_tokens: int,
        max_length: int,
        external_feature_extractor: bool,
        device: str,
        embeddings: Optional[nn.Embedding] = None,
        pre_computed_num_out_features: Optional[int] = None,
    ) -> None:
        super().__init__()
        self.model_config = model_config
        self.embedding_dim = embedding_dim
        self.num_tokens = num_tokens
        self.max_length = max_length
        self.external_feature_extractor = external_feature_extractor
        self.pre_computed_num_out_features = pre_computed_num_out_features

        pos_repr_class = get_positional_representation_class(
            position_model_config=self.model_config.position
        )
        self.pos_representation = pos_repr_class(
            embedding_dim=self.embedding_dim,
            dropout=self.model_config.position_dropout,
            max_length=self.max_length,
        )

        if embeddings:
            self.embedding = embeddings
        else:
            self.embedding = nn.Embedding(
                num_embeddings=self.num_tokens, embedding_dim=self.embedding_dim
            )

        self.feature_extractor = feature_extractor

        if not embeddings:
            self.init_embedding_weights()

        (
            self.dynamic_extras,
            self.extract_features,
        ) = _get_transformer_wrapper_feature_extractor(
            feature_extractor=self.feature_extractor,
            window_size=self.model_config.window_size,
            max_length=self.max_length,
            embedding_dim=self.embedding_dim,
            pool=self.model_config.pool,
            device=device,
            external_feature_extractor=self.external_feature_extractor,
        )

    @property
    def num_out_features(self) -> int:
        if self.pre_computed_num_out_features:
            return self.pre_computed_num_out_features

        padding = self.dynamic_extras.get("padding", 0)
        length_with_padding = self.max_length + padding

        if self.model_config.pool in ("avg", "max"):
            num_chunks = 1
            if self.model_config.window_size:
                num_chunks = length_with_padding // self.model_config.window_size

            return self.embedding_dim * num_chunks

        return length_with_padding * self.embedding_dim

    def init_embedding_weights(self) -> None:
        init_range = 0.1
        self.embedding.weight.data.uniform_(-init_range, init_range)

    def forward(self, input: Tensor) -> Tensor:
        out = input * math.sqrt(self.embedding_dim)
        out = self.pos_representation(out)
        out = self.extract_features(out)

        return out


def get_embedding_dim_for_sequence_model(
    embedding_dim: Union[None, int], num_tokens: int, num_heads: int
) -> int:
    if embedding_dim is None:
        auto_emb_dim = math.ceil((int(num_tokens**0.25) / num_heads)) * num_heads
        logger.info(
            "Setting up automatic embedding dimension of %d based on %d "
            "tokens and %d attention heads.",
            auto_emb_dim,
            num_tokens,
            num_heads,
        )
        embedding_dim = auto_emb_dim

    return embedding_dim


class SimpleFeatureExtractorForwardProtocol(Protocol):
    def __call__(
        self,
        input: torch.Tensor,
        feature_extractor: Union[nn.Module, "TransformerFeatureExtractor"],
    ) -> torch.Tensor: ...


class HFFeatureExtractorForwardProtocol(Protocol):
    def __call__(
        self,
        input: torch.Tensor,
        feature_extractor: Union[nn.Module, "TransformerFeatureExtractor"],
        key: str = "last_hidden_state",
    ) -> torch.Tensor: ...


def _get_transformer_wrapper_feature_extractor(
    feature_extractor: Union["TransformerFeatureExtractor", nn.Module],
    external_feature_extractor: bool,
    window_size: int,
    embedding_dim: int,
    max_length: int,
    device: str,
    pool: Union[Literal["avg"], Literal["max"], None] = None,
) -> Tuple[Dict[str, int], Callable[[torch.Tensor], torch.Tensor]]:
    dynamic_extras = {"padding": 0}

    feature_extractor_forward = _get_feature_extractor_forward(
        is_hf_model=external_feature_extractor,
        feature_extractor=feature_extractor,
        input_length=window_size if window_size else max_length,
        embedding_size=embedding_dim,
        device=device,
        pool=pool,
    )
    if not window_size:
        extractor = partial(
            feature_extractor_forward,
            feature_extractor=feature_extractor,
        )
    else:
        num_chunks = int(math.ceil(max_length / window_size))
        logger.debug(
            "Setting num chunks to %d as window size of %d and maximum sequence length "
            "of %d were passed in.",
            num_chunks,
            window_size,
            max_length,
        )

        padding = _find_lcl_padding_needed(
            input_size=max_length,
            kernel_size=window_size,
            num_chunks=num_chunks,
        )
        dynamic_extras["padding"] = padding

        extractor = partial(
            _conv_transformer_forward,
            feature_extractor=feature_extractor,
            feature_extractor_forward_callable=feature_extractor_forward,
            max_length=max_length,
            window_size=window_size,
            padding=padding,
        )

    return dynamic_extras, extractor


def _get_feature_extractor_forward(
    is_hf_model: bool,
    feature_extractor: Union[nn.Module, PreTrainedModel],
    input_length: int,
    embedding_size: int,
    device: str,
    pool: Union[Literal["avg"], Literal["max"], None] = None,
) -> SimpleFeatureExtractorForwardProtocol | HFFeatureExtractorForwardProtocol:
    if is_hf_model:
        return get_hf_transformer_forward(
            feature_extractor_=feature_extractor,
            input_length=input_length,
            embedding_dim=embedding_size,
            device=device,
            pool=pool,
        )
    return _get_simple_transformer_forward(pool=pool)


def _get_simple_transformer_forward(
    pool: Union[Literal["avg"], Literal["max"], None] = None,
):
    pooling_func = _get_sequence_pooling_func(pool=pool)

    def _simple_transformer_forward(
        input: torch.Tensor,
        feature_extractor: "TransformerFeatureExtractor",
    ) -> torch.Tensor:
        tensor_out = feature_extractor(input)
        tensor_pooled = pooling_func(input=tensor_out)
        final_out = tensor_pooled.flatten(1)

        return final_out

    return _simple_transformer_forward


def get_hf_transformer_forward(
    feature_extractor_: PreTrainedModel,
    input_length: int,
    embedding_dim: int,
    device: str,
    pool: Union[Literal["avg"], Literal["max"], None] = None,
):
    forward_argnames = inspect.getfullargspec(feature_extractor_.forward)[0]

    bound_kwargs = _build_transformer_forward_kwargs(
        forward_argnames=forward_argnames,
        config=feature_extractor_.config,
        input_length=input_length,
        embedding_dim=embedding_dim,
        device=device,
    )

    pooling_func = _get_sequence_pooling_func(pool=pool)

    def _hf_transformer_forward(
        input: torch.Tensor,
        feature_extractor: nn.Module,
        key: str = "last_hidden_state",
    ) -> torch.Tensor:
        hf_transformer_out = feature_extractor(inputs_embeds=input, **bound_kwargs)
        tensor_out = getattr(hf_transformer_out, key)
        tensor_pooled = pooling_func(input=tensor_out)
        final_out = tensor_pooled.flatten(1)
        return final_out

    return _hf_transformer_forward


def _get_sequence_pooling_func(
    pool: Union[Literal["avg"], Literal["max"], None]
) -> Callable:
    def _identity(input: torch.Tensor) -> torch.Tensor:
        return input

    def _max(input: torch.Tensor) -> torch.Tensor:
        return input.max(dim=1)[0]

    def _avg(input: torch.Tensor) -> torch.Tensor:
        return input.mean(dim=1)

    if pool is None:
        return _identity

    elif pool == "max":
        return _max

    elif pool == "avg":
        return _avg

    raise ValueError()


def _build_transformer_forward_kwargs(
    forward_argnames: Sequence[str],
    config: PretrainedConfig,
    input_length: int,
    embedding_dim: int,
    device: str,
) -> Dict:
    """
    TODO: Deprecate.
    """
    kwargs = {}

    if "decoder_inputs_embeds" in forward_argnames:
        kwargs["decoder_inputs_embeds"] = torch.randn(1, input_length, embedding_dim)

    return kwargs


def _conv_transformer_forward(
    input: torch.Tensor,
    feature_extractor: Union[nn.Module, "TransformerFeatureExtractor"],
    feature_extractor_forward_callable: (
        SimpleFeatureExtractorForwardProtocol | HFFeatureExtractorForwardProtocol
    ),
    max_length: int,
    window_size: int,
    padding: int,
) -> torch.Tensor:
    out = pad(input=input, pad=[0, 0, padding, 0])
    total_length = max_length + padding

    aggregated_out = None
    for lower_index in range(0, total_length, window_size):
        upper_index = lower_index + window_size

        cur_input = out[:, lower_index:upper_index, :]
        cur_out = feature_extractor_forward_callable(cur_input, feature_extractor)

        if aggregated_out is None:
            aggregated_out = cur_out
        else:
            aggregated_out = torch.cat((aggregated_out, cur_out), dim=1)

    assert aggregated_out is not None
    return aggregated_out


@dataclass
class BasicTransformerFeatureExtractorModelConfig:
    """
    :param num_heads:
        The number of heads in the multi-head attention models

    :param num_layers:
        The number of encoder blocks in the transformer model.

    :param dim_feedforward:
        The dimension of the feedforward layers in the transformer model.

    :param dropout:
         Dropout value to use in the encoder layers.
    """

    num_heads: int = 8
    num_layers: int = 2
    dim_feedforward: Union[int, Literal["auto"]] = "auto"
    dropout: float = 0.10


class TransformerFeatureExtractor(nn.Module):
    def __init__(
        self,
        model_config: BasicTransformerFeatureExtractorModelConfig,
        embedding_dim: int,
        num_tokens: int,
        max_length: int,
    ) -> None:
        super().__init__()

        self.model_config = model_config
        self.embedding_dim = embedding_dim
        self.num_tokens = num_tokens
        self.max_length = max_length

        dim_feed_forward = parse_dim_feedforward(
            dim_feedforward=self.model_config.dim_feedforward,
            embedding_dim=self.embedding_dim,
        )

        self.transformer_encoder = Transformer(
            d_model=self.embedding_dim,
            nhead=self.model_config.num_heads,
            num_layers=self.model_config.num_layers,
            dim_feedforward=dim_feed_forward,
            dropout=self.model_config.dropout,
            norm_first=True,
        )

    @property
    def num_out_features(self) -> int:
        return self.max_length * self.embedding_dim

    def forward(self, input: Tensor) -> Tensor:
        out = self.transformer_encoder(input)
        return out


def parse_dim_feedforward(
    dim_feedforward: Union[int, Literal["auto"]], embedding_dim: int
) -> int:
    if dim_feedforward == "auto":
        dim_feedforward = embedding_dim * 4
        logger.info(
            "Setting dim_feedfoward to %d based on embedding_dim=%d and 'auto' "
            "option.",
            dim_feedforward,
            embedding_dim,
        )

    return dim_feedforward


class SequenceOutputTransformerFeatureExtractor(TransformerFeatureExtractor):
    def __init__(
        self,
        model_config: BasicTransformerFeatureExtractorModelConfig,
        embedding_dim: int,
        num_tokens: int,
        max_length: int,
    ):
        super().__init__(
            model_config=model_config,
            embedding_dim=embedding_dim,
            num_tokens=num_tokens,
            max_length=max_length,
        )

        mask = torch.triu(
            torch.ones(self.max_length, self.max_length) * float("-inf"), diagonal=1
        )
        self.register_buffer("mask", mask)

    def forward(self, input: Tensor) -> Tensor:
        out = self.transformer_encoder(input, mask=self.mask)
        return out


def get_positional_representation_class(
    position_model_config: Literal["encode", "embed"]
) -> Union[Type["PositionalEncoding"], Type["PositionalEmbedding"]]:
    if position_model_config == "encode":
        return PositionalEncoding
    elif position_model_config == "embed":
        return PositionalEmbedding
    raise ValueError(
        "Unknown value for positional representation. "
        "Expected 'encode' or 'embed' but got '%s'.",
        position_model_config,
    )


class PositionalEncoding(nn.Module):
    def __init__(
        self,
        embedding_dim: int,
        max_length: int,
        dropout: float = 0.1,
    ) -> None:
        super().__init__()
        self.dropout = nn.Dropout(p=dropout)
        self.max_length = max_length

        position = torch.arange(max_length).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, embedding_dim, 2) * (-math.log(10000.0) / embedding_dim)
        )
        pe: torch.Tensor = torch.zeros(1, max_length, embedding_dim)
        pe[0, :, 0::2] = torch.sin(position * div_term)
        pe[0, :, 1::2] = torch.cos(position * div_term)
        self.register_buffer("pe", pe)

        self.pe: Tensor

    def forward(self, x: Tensor) -> Tensor:
        x = x + self.pe[:, : self.max_length, :]
        return self.dropout(x)


class PositionalEmbedding(nn.Module):
    def __init__(
        self,
        embedding_dim: int,
        max_length: int,
        dropout: float = 0.1,
        zero_init: bool = False,
    ) -> None:
        super().__init__()
        self.dropout = nn.Dropout(p=dropout)
        self.max_length = max_length

        if zero_init:
            self.embedding = torch.nn.Parameter(
                data=torch.zeros(1, max_length, embedding_dim),
                requires_grad=True,
            )
        else:
            self.embedding = torch.nn.Parameter(
                data=torch.randn(1, max_length, embedding_dim),
                requires_grad=True,
            )

    def forward(self, x: Tensor) -> Tensor:
        x = x + self.embedding[:, : self.max_length, :]
        return self.dropout(x)



from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Union

from eir.experiment_io.experiment_io import (
    LoadedTrainExperiment,
    load_serialized_train_experiment,
)
from eir.models import model_setup
from eir.models.meta.meta import MetaModel
from eir.serve_modules.serve_input_setup import set_up_inputs_for_serve
from eir.setup.config import Configs
from eir.setup.input_setup import al_input_objects_as_dict
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.train import Hooks, al_output_objects_as_dict
    from eir.train_utils.metrics import al_metric_record_dict


logger = get_logger(name=__name__, tqdm_compatible=True)


@dataclass(frozen=True)
class ServeExperiment:
    configs: Configs
    hooks: Union["Hooks", None]
    metrics: "al_metric_record_dict"
    inputs: al_input_objects_as_dict
    outputs: "al_output_objects_as_dict"
    model: MetaModel


def load_experiment_for_serve(
    model_path: str,
    device: str,
) -> ServeExperiment:
    model_path_object = Path(model_path)

    if not model_path_object.exists():
        raise FileNotFoundError(f"Model file not found at '{model_path}'")
    if not model_path_object.is_file():
        raise ValueError(f"Model path '{model_path}' is not a file.")
    if model_path_object.suffix != ".pt":
        raise ValueError(
            f"Model path '{model_path}' does not have the '.pt' extension as expected."
        )

    run_folder = model_path_object.parent.parent

    loaded_train_experiment = load_serialized_train_experiment(
        run_folder=run_folder, device=device
    )

    default_train_hooks = loaded_train_experiment.hooks
    train_configs = loaded_train_experiment.configs

    loaded_train_experiment.configs.gc.be.device = device

    inputs = set_up_inputs_for_serve(
        test_inputs_configs=train_configs.input_configs,
        hooks=default_train_hooks,
        output_folder=str(run_folder),
    )

    logger.info("Loading EIR PyTorch model from '%s'.", model_path)

    model = load_pytorch_eir_model_for_serve(
        model_pt_path=str(model_path),
        loaded_train_experiment=loaded_train_experiment,
        inputs=inputs,
        device=device,
    )

    assert not model.training

    loaded_train_experiment_as_dict = loaded_train_experiment.__dict__
    serve_experiment_kwargs = {
        **{"model": model, "inputs": inputs},
        **loaded_train_experiment_as_dict,
    }

    serve_experiment = ServeExperiment(**serve_experiment_kwargs)

    return serve_experiment


def load_pytorch_eir_model_for_serve(
    model_pt_path: str,
    loaded_train_experiment: LoadedTrainExperiment,
    inputs: al_input_objects_as_dict,
    device: str,
) -> model_setup.al_meta_model:
    func = model_setup.get_meta_model_class_and_kwargs_from_configs
    meta_model_class, meta_model_kwargs = func(
        global_config=loaded_train_experiment.configs.global_config,
        fusion_config=loaded_train_experiment.configs.fusion_config,
        inputs_as_dict=inputs,
        outputs_as_dict=loaded_train_experiment.outputs,
    )

    model = model_setup.load_model(
        model_path=Path(model_pt_path),
        model_class=meta_model_class,
        model_init_kwargs=meta_model_kwargs,
        device=device,
        test_mode=True,
        strict_shapes=True,
    )
    assert not model.training

    for name, _ in model.named_modules():
        if "class" in name.split("."):
            raise NameError(
                f"'class' is a reserved keyword and cannot be part of the "
                f"modules in model. Please ensure no module is named 'class'. "
                f"Found in '{name}'."
            )

    return model
